# COPYRIGHT (C) 2020-2022 Nicotine+ Team
# COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
# COPYRIGHT (C) 2008-2010 Quinox <quinox@users.sf.net>
# COPYRIGHT (C) 2006-2009 Daelstorm <daelstorm@gmail.com>
# COPYRIGHT (C) 2003-2004 Hyriand <hyriand@thegraveyard.org>
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import time

from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import Gtk

from pynicotine.config import config
from pynicotine.geoip.geoip import GeoIP
from pynicotine.gtkgui.widgets.filechooser import save_file
from pynicotine.gtkgui.widgets.iconnotebook import IconNotebook
from pynicotine.gtkgui.widgets.infobar import InfoBar
from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
from pynicotine.gtkgui.widgets.textview import TextView
from pynicotine.gtkgui.widgets.theme import update_widget_visuals
from pynicotine.gtkgui.widgets.treeview import initialise_columns
from pynicotine.gtkgui.widgets.ui import UserInterface
from pynicotine.logfacility import log
from pynicotine.utils import humanize
from pynicotine.utils import human_speed


class UserInfos(IconNotebook):

    def __init__(self, frame):

        IconNotebook.__init__(self, frame, frame.userinfo_notebook, "userinfo")
        self.notebook.connect("switch-page", self.on_switch_info_page)

    def on_switch_info_page(self, _notebook, page, _page_num):

        if self.frame.current_page_id != self.page_id:
            return

        for tab in self.pages.values():
            if tab.Main == page:
                GLib.idle_add(lambda: tab.descr.grab_focus() == -1)  # pylint:disable=cell-var-from-loop
                break

    def show_user(self, user, switch_page=True):

        if user not in self.pages:
            self.pages[user] = page = UserInfo(self, user)
            self.append_page(page.Main, user, page.on_close, user=user)
            page.set_label(self.get_tab_label_inner(page.Main))

        if switch_page:
            self.set_current_page(self.page_num(self.pages[user].Main))
            self.frame.change_main_page("userinfo")

    def show_connection_error(self, user):
        if user in self.pages:
            self.pages[user].show_connection_error()

    def message_progress(self, msg):
        if msg.user in self.pages:
            self.pages[msg.user].message_progress(msg)

    def get_user_stats(self, msg):
        if msg.user in self.pages:
            self.pages[msg.user].get_user_stats(msg)

    def get_user_status(self, msg):

        if msg.user in self.pages:
            page = self.pages[msg.user]
            self.set_user_status(page.Main, msg.user, msg.status)

    def set_user_country(self, user, country_code):
        if user in self.pages:
            self.pages[user].set_user_country(country_code)

    def user_interests(self, msg):
        if msg.user in self.pages:
            self.pages[msg.user].user_interests(msg)

    def user_info_reply(self, user, msg):
        if user in self.pages:
            self.pages[user].user_info_reply(msg)

    def update_visuals(self):
        for page in self.pages.values():
            page.update_visuals()

    def server_disconnect(self):
        for user, page in self.pages.items():
            self.set_user_status(page.Main, user, 0)


class UserInfo(UserInterface):

    def __init__(self, userinfos, user):

        super().__init__("ui/userinfo.ui")

        self.userinfos = userinfos
        self.frame = userinfos.frame

        self.info_bar = InfoBar(self.InfoBar, Gtk.MessageType.INFO)
        self.descr_textview = TextView(self.descr)
        self.UserLabel.set_text(user)

        if Gtk.get_major_version() == 4:
            self.picture = Gtk.Picture(can_shrink=False, halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER)
            self.picture_view.set_child(self.picture)

            self.scroll_controller = Gtk.EventControllerScroll(flags=Gtk.EventControllerScrollFlags.VERTICAL)
            self.scroll_controller.connect("scroll", self.on_scroll)
            self.picture_view.add_controller(self.scroll_controller)

        else:
            self.picture = Gtk.Image(visible=True)
            self.picture_view.add(self.picture)
            self.picture_view.connect("scroll-event", self.on_scroll_event)

        self.user = user
        self.picture_data = None
        self.zoom_factor = 5
        self.actual_zoom = 0

        # Set up likes list
        self.likes_store = Gtk.ListStore(str)

        self.like_column_numbers = list(range(self.likes_store.get_n_columns()))
        cols = initialise_columns(
            self.frame, None, self.Likes,
            ["likes", _("Likes"), 0, "text", None]
        )
        cols["likes"].set_sort_column_id(0)

        self.likes_store.set_sort_column_id(0, Gtk.SortType.ASCENDING)
        self.Likes.set_model(self.likes_store)

        # Set up dislikes list
        self.hates_store = Gtk.ListStore(str)

        self.hate_column_numbers = list(range(self.hates_store.get_n_columns()))
        cols = initialise_columns(
            self.frame, None, self.Hates,
            ["dislikes", _("Dislikes"), 0, "text", None]
        )
        cols["dislikes"].set_sort_column_id(0)

        self.hates_store.set_sort_column_id(0, Gtk.SortType.ASCENDING)
        self.Hates.set_model(self.hates_store)

        # Popup menus
        self.user_popup = popup = PopupMenu(self.frame, None, self.on_tab_popup)
        popup.setup_user_menu(user, page="userinfo")
        popup.add_items(
            ("", None),
            ("#" + _("Close All Tabs…"), self.on_close_all_tabs),
            ("#" + _("_Close Tab"), self.on_close)
        )

        def get_interest_items(popup):
            return (("$" + _("I _Like This"), self.on_like_recommendation, popup),
                    ("$" + _("I _Dislike This"), self.on_dislike_recommendation, popup),
                    ("", None),
                    ("#" + _("_Search for Item"), self.on_interest_recommend_search, popup))

        popup = PopupMenu(self.frame, self.Likes, self.on_popup_interest_menu)
        popup.add_items(*get_interest_items(popup))

        popup = PopupMenu(self.frame, self.Hates, self.on_popup_interest_menu)
        popup.add_items(*get_interest_items(popup))

        popup = PopupMenu(self.frame, self.picture_view)
        popup.add_items(
            ("#" + _("Zoom 1:1"), self.make_zoom_normal),
            ("#" + _("Zoom In"), self.make_zoom_in),
            ("#" + _("Zoom Out"), self.make_zoom_out),
            ("", None),
            ("#" + _("Save Picture"), self.on_save_picture)
        )

        self.update_visuals()

    def set_label(self, label):
        self.user_popup.set_parent(label)

    def save_columns(self):
        # Unused
        pass

    def update_visuals(self):

        for widget in list(self.__dict__.values()):
            update_widget_visuals(widget)

    """ General """

    def load_picture(self, data):

        if not data:
            self.picture_view.hide()
            return

        try:
            import gc
            import tempfile

            with tempfile.NamedTemporaryFile() as file_handle:
                file_handle.write(data)
                del data

                self.picture_data = GdkPixbuf.Pixbuf.new_from_file(file_handle.name)
                picture_width = self.picture_data.get_width()
                picture_height = self.picture_data.get_height()

                allocation = self.placeholder_picture.get_allocation()
                max_width = allocation.width - 72
                max_height = allocation.height - 72

                # Resize picture to fit container
                ratio = min(max_width / picture_width, max_height / picture_height)
                self.picture_data = self.picture_data.scale_simple(
                    ratio * picture_width, ratio * picture_height, GdkPixbuf.InterpType.BILINEAR)

                if Gtk.get_major_version() == 4:
                    self.picture.set_pixbuf(self.picture_data)
                else:
                    self.picture.set_from_pixbuf(self.picture_data)

            gc.collect()

            self.actual_zoom = 0
            self.picture_view.show()

        except Exception as error:
            log.add(_("Failed to load picture for user %(user)s: %(error)s"), {
                "user": self.user,
                "error": error
            })

    def make_zoom_normal(self, *_args):
        self.make_zoom_in(zoom=True)

    def make_zoom_in(self, *_args, zoom=None):

        def calc_zoom_in(w_h):
            return w_h + w_h * self.actual_zoom / 100 + w_h * self.zoom_factor / 100

        import gc

        if self.picture is None or self.picture_data is None or self.actual_zoom > 100:
            return

        width = self.picture_data.get_width()
        height = self.picture_data.get_height()

        if zoom:
            self.actual_zoom = 0
            picture_zoomed = self.picture_data

        else:
            self.actual_zoom += self.zoom_factor
            picture_zoomed = self.picture_data.scale_simple(
                calc_zoom_in(width), calc_zoom_in(height), GdkPixbuf.InterpType.BILINEAR)

        if Gtk.get_major_version() == 4:
            self.picture.set_pixbuf(picture_zoomed)
        else:
            self.picture.set_from_pixbuf(picture_zoomed)

        del picture_zoomed
        gc.collect()

    def make_zoom_out(self, *_args):

        def calc_zoom_out(w_h):
            return w_h + w_h * self.actual_zoom / 100 - w_h * self.zoom_factor / 100

        import gc

        if self.picture is None or self.picture_data is None:
            return

        width = self.picture_data.get_width()
        height = self.picture_data.get_height()

        self.actual_zoom -= self.zoom_factor

        if calc_zoom_out(width) < 10 or calc_zoom_out(height) < 10:
            self.actual_zoom += self.zoom_factor
            return

        picture_zoomed = self.picture_data.scale_simple(
            calc_zoom_out(width), calc_zoom_out(height), GdkPixbuf.InterpType.BILINEAR)

        if Gtk.get_major_version() == 4:
            self.picture.set_pixbuf(picture_zoomed)
        else:
            self.picture.set_from_pixbuf(picture_zoomed)

        del picture_zoomed
        gc.collect()

    def show_connection_error(self):

        self.info_bar.show_message(
            _("Unable to request information from user. Either you both have a closed listening "
              "port, the user is offline, or there's a temporary connectivity issue.")
        )

        self.set_finished()

    def set_finished(self):
        self.userinfos.request_tab_hilite(self.Main)
        self.progressbar.set_fraction(1.0)

    def message_progress(self, msg):

        if msg.total == 0 or msg.position == 0:
            fraction = 0.0
        elif msg.position >= msg.total:
            fraction = 1.0
        else:
            fraction = float(msg.position) / msg.total

        self.progressbar.set_fraction(fraction)

    """ Network Messages """

    def user_info_reply(self, msg):

        if msg is None:
            return

        if msg.descr:
            self.descr_textview.clear()
            self.descr_textview.append_line(msg.descr, showstamp=False, scroll=False)

        self.uploads.set_text(humanize(msg.totalupl))
        self.queuesize.set_text(humanize(msg.queuesize))
        self.slotsavail.set_text(_("Yes") if msg.slotsavail else _("No"))

        self.picture_data = None
        self.load_picture(msg.pic)

        self.info_bar.set_visible(False)
        self.set_finished()

    def get_user_stats(self, msg):

        if msg.avgspeed > 0:
            self.speed.set_text(human_speed(msg.avgspeed))

        self.filesshared.set_text(humanize(msg.files))
        self.dirsshared.set_text(humanize(msg.dirs))

    def set_user_country(self, country_code):

        if country_code:
            country = GeoIP.country_code_to_name(country_code)
            country_text = "%s (%s)" % (country, country_code)
        else:
            country_text = _("Unknown")

        self.country.set_text(country_text)

    def user_interests(self, msg):

        self.likes_store.clear()
        self.hates_store.clear()

        for like in msg.likes:
            self.likes_store.insert_with_valuesv(-1, self.like_column_numbers, [like])

        for hate in msg.hates:
            self.hates_store.insert_with_valuesv(-1, self.hate_column_numbers, [hate])

    """ Callbacks """

    def on_tab_popup(self, *_args):
        self.user_popup.toggle_user_items()

    @staticmethod
    def on_popup_interest_menu(menu, widget):

        model, iterator = widget.get_selection().get_selected()
        item = model.get_value(iterator, 0)
        menu.set_user(item)

        menu.actions[_("I _Like This")].set_state(
            GLib.Variant("b", item in config.sections["interests"]["likes"])
        )
        menu.actions[_("I _Dislike This")].set_state(
            GLib.Variant("b", item in config.sections["interests"]["dislikes"])
        )

    def on_like_recommendation(self, action, state, popup):
        self.frame.interests.on_like_recommendation(action, state, popup.get_user())

    def on_dislike_recommendation(self, action, state, popup):
        self.frame.interests.on_dislike_recommendation(action, state, popup.get_user())

    def on_interest_recommend_search(self, _action, _state, popup):
        self.frame.interests.recommend_search(popup.get_user())

    def on_send_message(self, *_args):
        self.frame.np.privatechats.show_user(self.user)
        self.frame.change_main_page("private")

    def on_show_ip_address(self, *_args):
        self.frame.np.request_ip_address(self.user)

    def on_browse_user(self, *_args):
        self.frame.np.userbrowse.browse_user(self.user)

    def on_add_to_list(self, *_args):
        self.frame.np.userlist.add_user(self.user)

    def on_ban_user(self, *_args):
        self.frame.np.network_filter.ban_user(self.user)

    def on_ignore_user(self, *_args):
        self.frame.np.network_filter.ignore_user(self.user)

    def on_save_picture_response(self, selected, _data):

        if not os.path.exists(selected):
            self.picture_data.savev(selected, "jpeg", ["quality"], ["100"])
            log.add(_("Picture saved to %s"), selected)
        else:
            log.add(_("Picture not saved, %s already exists."), selected)

    def on_save_picture(self, *_args):

        if self.picture is None or self.picture_data is None:
            return

        save_file(
            parent=self.frame.MainWindow,
            callback=self.on_save_picture_response,
            initialdir=config.sections["transfers"]["downloaddir"],
            initialfile="%s %s.jpg" % (self.user, time.strftime("%Y-%m-%d %H_%M_%S")),
            title=_("Save as…")
        )

    def on_scroll(self, _controller, _scroll_x, scroll_y):

        if scroll_y < 0:
            self.make_zoom_in()
        else:
            self.make_zoom_out()

        return True

    def on_scroll_event(self, _widget, event):

        if event.get_scroll_deltas().delta_y < 0:
            self.make_zoom_in()
        else:
            self.make_zoom_out()

        return True

    def on_refresh(self, *_args):

        self.info_bar.set_visible(False)
        self.progressbar.set_fraction(0.0)

        self.frame.np.userinfo.request_user_info(self.user)

    def on_close(self, *_args):

        del self.userinfos.pages[self.user]
        self.frame.np.userinfo.remove_user(self.user)
        self.userinfos.remove_page(self.Main)

    def on_close_all_tabs(self, *_args):
        self.userinfos.remove_all_pages()



# COPYRIGHT (C) 2021-2022 Nicotine+ Team
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os


def check_gui_dependencies():

    if os.getenv("NICOTINE_GTK_VERSION") == '4':
        gtk_version = (4, 6, 0)
        pygobject_version = (3, 40, 0)
    else:
        gtk_version = pygobject_version = (3, 18, 0)

    try:
        import gi
        gi.check_version(pygobject_version)

    except (ImportError, ValueError):
        return _("Cannot find %s, please install it.") % ("pygobject >= " + '.'.join(map(str, pygobject_version)))

    try:
        api_version = (gtk_version[0], 0)
        gi.require_version('Gtk', '.'.join(map(str, api_version)))

    except ValueError:
        return _("Cannot find %s, please install it.") % ("GTK " + str(gtk_version[0]))

    try:
        from gi.repository import Gtk

    except ImportError:
        return _("Cannot import the Gtk module. Bad install of the python-gobject module?")

    if Gtk.check_version(*gtk_version):
        return _("You are using an unsupported version of GTK %(major_version)s. You should install "
                 "GTK %(complete_version)s or newer.") % {
            "major_version": gtk_version[0],
            "complete_version": '.'.join(map(str, gtk_version))}

    try:
        if gtk_version[0] == 4 and os.getenv("NICOTINE_LIBADWAITA") == '1':
            gi.require_version('Adw', '1')

            from gi.repository import Adw
            Adw.init()

    except (ImportError, ValueError):
        pass

    return None


def run_gui(network_processor, trayicon, hidden, bindip, port, ci_mode, multi_instance):
    """ Run Nicotine+ GTK GUI """

    from pynicotine.logfacility import log
    error = check_gui_dependencies()

    if error:
        log.add(error)
        return 1

    from pynicotine.gtkgui.frame import Application
    return Application(network_processor, trayicon, hidden, bindip, port, ci_mode, multi_instance).run()



# COPYRIGHT (C) 2020-2022 Nicotine+ Team
# COPYRIGHT (C) 2009 Quinox <quinox@users.sf.net>
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pynicotine.pluginsystem import BasePlugin


class Plugin(BasePlugin):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.log('__init__()')

    def init(self):
        self.log('init()')

    def disable(self):
        self.log('disable()')

    def loaded_notification(self):
        self.log('loaded_notification()')

    def unloaded_notification(self):
        self.log('unloaded_notification()')

    def shutdown_notification(self):
        self.log('shutdown_notification()')

    def public_room_message_notification(self, room, user, line):
        self.log('public_room_message_notification(room=%s, user=%s, line=%s)', (room, user, line))

    def search_request_notification(self, searchterm, user, token):
        self.log('search_request_notification(searchterm=%s, user=%s, token=%s)', (searchterm, user, token))

    def distrib_search_notification(self, searchterm, user, token):
        # Verbose:
        # self.log('distrib_search_notification(searchterm=%s, user=%s, token=%s)', (searchterm, user, token))
        pass

    def incoming_private_chat_event(self, user, line):
        self.log('incoming_private_chat_event(user=%s, line=%s)', (user, line))

    def incoming_private_chat_notification(self, user, line):
        self.log('incoming_private_chat_notification(user=%s, line=%s)', (user, line))

    def incoming_public_chat_event(self, room, user, line):
        self.log('incoming_public_chat_event(room=%s, user=%s, line=%s)', (room, user, line))

    def incoming_public_chat_notification(self, room, user, line):
        self.log('incoming_public_chat_notification(room=%s, user=%s, line=%s)', (room, user, line))

    def outgoing_private_chat_event(self, user, line):
        self.log('outgoing_private_chat_event(user=%s, line=%s)', (user, line))

    def outgoing_private_chat_notification(self, user, line):
        self.log('outgoing_private_chat_notification(user=%s, line=%s)', (user, line))

    def outgoing_public_chat_event(self, room, line):
        self.log('outgoing_public_chat_event(room=%s, line=%s)', (room, line))

    def outgoing_public_chat_notification(self, room, line):
        self.log('outgoing_public_chat_notification(room=%s, line=%s)', (room, line))

    def outgoing_global_search_event(self, text):
        self.log('outgoing_global_search_event(text=%s)', (text,))

    def outgoing_room_search_event(self, rooms, text):
        self.log('outgoing_room_search_event(rooms=%s, text=%s)', (rooms, text))

    def outgoing_buddy_search_event(self, text):
        self.log('outgoing_buddy_search_event(text=%s)', (text,))

    def outgoing_user_search_event(self, users, text):
        self.log('outgoing_user_search_event(users=%s, text=%s)', (users, text))

    def user_resolve_notification(self, user, ip_address, port, country):
        self.log('user_resolve_notification(user=%s, ip_address=%s, port=%s, country=%s)',
                 (user, ip_address, port, country))

    def server_connect_notification(self):
        self.log('server_connect_notification()')

    def server_disconnect_notification(self, userchoice):
        self.log('server_disconnect_notification(userchoice=%s)', (userchoice,))

    def join_chatroom_notification(self, room):
        self.log('join_chatroom_notification(room=%s)', (room,))

    def leave_chatroom_notification(self, room):
        self.log('leave_chatroom_notification(room=%s)', (room,))

    def user_join_chatroom_notification(self, room, user):
        self.log('user_join_chatroom_notification(room=%s, user=%s)', (room, user,))

    def user_leave_chatroom_notification(self, room, user):
        self.log('user_leave_chatroom_notification(room=%s, user=%s)', (room, user,))

    def user_stats_notification(self, user, stats):
        self.log('user_stats_notification(user=%s, stats=%s)', (user, stats))

    def user_status_notification(self, user, status, privileged):
        self.log('user_status_notification(user=%s, status=%s, privileged=%s)', (user, status, privileged))

    def upload_queued_notification(self, user, virtual_path, real_path):
        self.log('upload_queued_notification(user=%s, virtual_path=%s, real_path=%s)',
                 (user, virtual_path, real_path))

    def upload_started_notification(self, user, virtual_path, real_path):
        self.log('upload_started_notification(user=%s, virtual_path=%s, real_path=%s)',
                 (user, virtual_path, real_path))

    def upload_finished_notification(self, user, virtual_path, real_path):
        self.log('upload_finished_notification(user=%s, virtual_path=%s, real_path=%s)',
                 (user, virtual_path, real_path))

    def download_started_notification(self, user, virtual_path, real_path):
        self.log('download_started_notification(user=%s, virtual_path=%s, real_path=%s)',
                 (user, virtual_path, real_path))

    def download_finished_notification(self, user, virtual_path, real_path):
        self.log('download_finished_notification(user=%s, virtual_path=%s, real_path=%s)',
                 (user, virtual_path, real_path))



# COPYRIGHT (C) 2020-2022 Nicotine+ Team
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk
from gi.repository import Gtk

from pynicotine.config import config

""" Dialogs """


def generic_dialog(parent=None, content_box=None, quit_callback=None,
                   title="Dialog", width=400, height=400, modal=True):

    dialog = Gtk.Dialog(
        use_header_bar=config.sections["ui"]["header_bar"],
        title=title,
        default_width=width,
        default_height=height
    )

    if content_box:
        dialog.get_content_area().add(content_box)

    set_dialog_properties(dialog, parent, quit_callback, modal)
    return dialog


def set_dialog_properties(dialog, parent, quit_callback=None, modal=True):

    if Gtk.get_major_version() == 4:
        if quit_callback:
            dialog.connect("close-request", quit_callback)
    else:
        dialog.set_property("window-position", Gtk.WindowPosition.CENTER_ON_PARENT)
        dialog.set_type_hint(Gdk.WindowTypeHint.DIALOG)

        if quit_callback:
            dialog.connect("delete-event", quit_callback)

        if isinstance(dialog, Gtk.Dialog):
            content_area = dialog.get_content_area()
            content_area.set_border_width(0)

    dialog.set_modal(modal)
    dialog.set_transient_for(parent)


def dialog_show(dialog):

    parent = dialog.get_transient_for()

    # Shrink the dialog if it's larger than the main window
    if Gtk.get_major_version() == 4:
        main_width = parent.get_width()
        main_height = parent.get_height()
    else:
        main_width, main_height = parent.get_size()

    new_width = dialog_width = dialog.get_property("default-width")
    new_height = dialog_height = dialog.get_property("default-height")

    if dialog_width > main_width:
        new_width = main_width - 30

    if dialog_height > main_height:
        new_height = main_height - 30

    if new_width > 0 and new_height > 0:
        if Gtk.get_major_version() == 4:
            dialog.set_default_size(new_width, new_height)
        else:
            dialog.resize(new_width, new_height)

    # Show the dialog
    dialog.present_with_time(Gdk.CURRENT_TIME)

    if Gtk.get_major_version() == 3:
        dialog.get_window().set_functions(
            Gdk.WMFunction.RESIZE | Gdk.WMFunction.MOVE | Gdk.WMFunction.CLOSE
        )


def dialog_hide(dialog):

    # Hide the dialog
    dialog.hide()

    # "Soft-delete" the dialog. This is necessary to prevent the dialog from
    # appearing in window peek on Windows
    dialog.unrealize()


""" Message Dialogs """


def entry_dialog(parent, title, message, callback, callback_data=None, default="",
                 optionmessage="", optionvalue=False, visibility=True, droplist=None):

    self = Gtk.MessageDialog(
        transient_for=parent,
        default_width=500,
        message_type=Gtk.MessageType.OTHER,
        buttons=Gtk.ButtonsType.OK_CANCEL,
        destroy_with_parent=True,
        modal=True,
        text=title,
        secondary_text=message
    )
    self.connect("response", callback, callback_data)

    if Gtk.get_major_version() == 4:
        label = self.get_message_area().get_last_child()
    else:
        label = self.get_message_area().get_children()[-1]

    label.set_selectable(True)

    if droplist:
        dropdown = Gtk.ComboBoxText(has_entry=True, visible=True)
        entry = dropdown.get_child()

        for i in droplist:
            dropdown.append_text(i)

        self.get_message_area().add(dropdown)
    else:
        entry = Gtk.Entry(visible=True)
        self.get_message_area().add(entry)

    self.get_response_value = entry.get_text
    entry.connect("activate", lambda x: self.response(Gtk.ResponseType.OK))
    entry.set_activates_default(True)
    entry.set_text(default)
    entry.set_visibility(visibility)

    if optionmessage:
        self.option = Gtk.CheckButton(label=optionmessage, active=optionvalue, visible=True)
        self.get_message_area().add(self.option)
        self.get_second_response_value = self.option.get_active

    self.present_with_time(Gdk.CURRENT_TIME)


def message_dialog(parent, title, message, callback=None):

    self = Gtk.MessageDialog(
        transient_for=parent,
        message_type=Gtk.MessageType.INFO,
        buttons=Gtk.ButtonsType.OK,
        destroy_with_parent=True,
        modal=True,
        text=title,
        secondary_text=message
    )

    if not callback:
        def callback(dialog, *_args):
            dialog.destroy()

    self.connect("response", callback)

    if Gtk.get_major_version() == 4:
        label = self.get_message_area().get_last_child()
    else:
        label = self.get_message_area().get_children()[-1]

    label.set_selectable(True)

    self.present_with_time(Gdk.CURRENT_TIME)


def option_dialog(parent, title, message, callback, callback_data=None, checkbox_label="",
                  first_button=_("_No"), second_button=_("_Yes"), third_button=""):

    self = Gtk.MessageDialog(
        transient_for=parent,
        message_type=Gtk.MessageType.QUESTION,
        destroy_with_parent=True,
        modal=True,
        text=title,
        secondary_text=message
    )
    self.connect("response", callback, callback_data)

    if Gtk.get_major_version() == 4:
        label = self.get_message_area().get_last_child()
    else:
        label = self.get_message_area().get_children()[-1]

    label.set_selectable(True)

    self.checkbox = Gtk.CheckButton(label=checkbox_label, visible=bool(checkbox_label))

    if checkbox_label:
        self.get_message_area().add(self.checkbox)

    if first_button:
        self.add_button(first_button, 1)

    if second_button:
        self.add_button(second_button, 2)

    if third_button:
        self.add_button(third_button, 3)

    self.present_with_time(Gdk.CURRENT_TIME)



#!/usr/bin/env python3
# COPYRIGHT (C) 2020-2022 Nicotine+ Team
# COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
# COPYRIGHT (C) 2009-2010 Quinox <quinox@users.sf.net>
# COPYRIGHT (C) 2009 Hedonist <ak@sensi.org>
# COPYRIGHT (C) 2006-2009 Daelstorm <daelstorm@gmail.com>
# COPYRIGHT (C) 2008-2009 eL_vErDe <gandalf@le-vert.net>
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
    To install Nicotine+ on a GNU/Linux distribution, run:
    pip3 install .
"""

import glob

from pkgutil import walk_packages
from setuptools import setup
from setuptools.command.build_py import build_py

import pynicotine

from pynicotine.config import config
from pynicotine.i18n import build_translations
from pynicotine.i18n import get_translation_paths


class BuildPyCommand(build_py):

    def run(self):
        build_translations()
        build_py.run(self)


if __name__ == '__main__':

    # Specify a description for the PyPi project page
    LONG_DESCRIPTION = """Nicotine+ is a graphical client for the Soulseek peer-to-peer
network.

Nicotine+ aims to be a pleasant, free and open source (FOSS)
alternative to the official Soulseek client, providing additional
functionality while keeping current with the Soulseek protocol."""

    # Specify included files
    PACKAGES = ["pynicotine"] + \
        [name for importer, name, ispkg in walk_packages(path=pynicotine.__path__, prefix="pynicotine.") if ispkg]

    PACKAGE_DATA = {package: ["*.bin", "*.md", "*.py", "*.svg", "*.ui", "PLUGININFO"] for package in PACKAGES}

    DATA_FILES = [
        ("share/applications", ["data/" + config.application_id + ".desktop"]),
        ("share/metainfo", ["data/" + config.application_id + ".metainfo.xml"]),
        ("share/icons/hicolor/scalable/apps", glob.glob("pynicotine/gtkgui/icons/hicolor/scalable/apps/*.svg")),
        ("share/icons/hicolor/scalable/intl", glob.glob("pynicotine/gtkgui/icons/hicolor/scalable/intl/*.svg")),
        ("share/icons/hicolor/symbolic/apps", glob.glob("pynicotine/gtkgui/icons/hicolor/symbolic/apps/*.svg")),
        ("share/doc/nicotine", glob.glob("[!404.md]*.md") + glob.glob("doc/*.md")),
        ("share/man/man1", glob.glob("data/*.1"))
    ] + get_translation_paths()

    # Run setup
    setup(
        name="nicotine-plus",
        version="3.2.1.post1",
        license="GPLv3+",
        description="Graphical client for the Soulseek peer-to-peer network",
        long_description=LONG_DESCRIPTION,
        author="Nicotine+ Team",
        author_email="nicotine-team@lists.launchpad.net",
        url=config.website_url,
        platforms="any",
        packages=PACKAGES,
        package_data=PACKAGE_DATA,
        scripts=["nicotine"],
        data_files=DATA_FILES,
        python_requires=">=3.5",
        install_requires=["PyGObject>=3.18"],
        cmdclass={"build_py": BuildPyCommand}
    )



# COPYRIGHT (C) 2020-2022 Nicotine+ Team
# COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
# COPYRIGHT (C) 2016 Mutnick <muhing@yahoo.com>
# COPYRIGHT (C) 2013 eL_vErDe <gandalf@le-vert.net>
# COPYRIGHT (C) 2008-2012 Quinox <quinox@users.sf.net>
# COPYRIGHT (C) 2009 Hedonist <ak@sensi.org>
# COPYRIGHT (C) 2006-2009 Daelstorm <daelstorm@gmail.com>
# COPYRIGHT (C) 2003-2004 Hyriand <hyriand@thegraveyard.org>
# COPYRIGHT (C) 2001-2003 Alexander Kanavin
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This module contains classes that deal with file transfers:
the transfer manager.
"""

import json
import os
import os.path
import re
import stat
import threading
import time

from collections import defaultdict
from collections import deque
from collections import OrderedDict

from pynicotine import slskmessages
from pynicotine.logfacility import log
from pynicotine.slskmessages import increment_token
from pynicotine.utils import execute_command
from pynicotine.utils import clean_file
from pynicotine.utils import clean_path
from pynicotine.utils import get_result_bitrate_length
from pynicotine.utils import human_length
from pynicotine.utils import load_file
from pynicotine.utils import write_file_and_backup


class Transfer:
    """ This class holds information about a single transfer """

    __slots__ = ("sock", "user", "filename",
                 "path", "token", "size", "file", "start_time", "last_update",
                 "current_byte_offset", "last_byte_offset", "speed", "time_elapsed",
                 "time_left", "modifier", "queue_position", "bitrate", "length",
                 "iterator", "status", "legacy_attempt")

    def __init__(self, user=None, filename=None, path=None, status=None, token=None, size=0,
                 current_byte_offset=None, bitrate=None, length=None):
        self.user = user
        self.filename = filename
        self.path = path
        self.size = size
        self.status = status
        self.token = token
        self.current_byte_offset = current_byte_offset
        self.bitrate = bitrate
        self.length = length

        self.sock = None
        self.file = None
        self.queue_position = 0
        self.modifier = None
        self.start_time = None
        self.last_update = None
        self.last_byte_offset = None
        self.speed = None
        self.time_elapsed = None
        self.time_left = None
        self.iterator = None
        self.legacy_attempt = False


class Transfers:
    """ This is the transfers manager """

    def __init__(self, core, config, queue, network_callback, ui_callback=None):

        self.core = core
        self.config = config
        self.queue = queue
        self.allow_saving_transfers = False
        self.downloads = deque()
        self.uploads = deque()
        self.privileged_users = set()
        self.requested_folders = defaultdict(dict)
        self.last_save_times = {"downloads": 0, "uploads": 0}
        self.transfer_request_times = {}
        self.upload_speed = 0
        self.token = 0

        self.user_update_counter = 0
        self.user_update_counters = {}

        self.downloads_file_name = os.path.join(self.config.data_dir, 'downloads.json')
        self.uploads_file_name = os.path.join(self.config.data_dir, 'uploads.json')

        self.network_callback = network_callback
        self.download_queue_timer_count = -1
        self.downloadsview = None
        self.uploadsview = None

        if hasattr(ui_callback, "downloads"):
            self.downloadsview = ui_callback.downloads

        if hasattr(ui_callback, "uploads"):
            self.uploadsview = ui_callback.uploads

        self.update_download_filters()

    def init_transfers(self):

        self.add_stored_transfers("downloads")
        self.add_stored_transfers("uploads")

        if self.downloadsview:
            self.downloadsview.init_transfers(self.downloads)

        if self.uploadsview:
            self.uploadsview.init_transfers(self.uploads)

        self.allow_saving_transfers = True

    def server_login(self):

        self.requested_folders.clear()
        self.update_limits()
        self.watch_stored_downloads()

        # Check for transfer timeouts
        thread = threading.Thread(target=self._check_transfer_timeouts)
        thread.name = "TransferTimeoutTimer"
        thread.daemon = True
        thread.start()

        # Check for failed downloads (1 min delay)
        thread = threading.Thread(target=self._check_download_queue_timer)
        thread.name = "DownloadQueueTimer"
        thread.daemon = True
        thread.start()

        # Check if queued uploads can be started
        thread = threading.Thread(target=self._check_upload_queue_timer)
        thread.name = "UploadQueueTimer"
        thread.daemon = True
        thread.start()

        if self.downloadsview:
            self.downloadsview.server_login()

        if self.uploadsview:
            self.uploadsview.server_login()

    """ Load Transfers """

    def get_download_queue_file_name(self):

        data_dir = self.config.data_dir
        downloads_file_json = os.path.join(data_dir, 'downloads.json')
        downloads_file_1_4_2 = os.path.join(data_dir, 'config.transfers.pickle')
        downloads_file_1_4_1 = os.path.join(data_dir, 'transfers.pickle')

        if os.path.exists(downloads_file_json):
            # New file format
            return downloads_file_json

        if os.path.exists(downloads_file_1_4_2):
            # Nicotine+ 1.4.2+
            return downloads_file_1_4_2

        if os.path.exists(downloads_file_1_4_1):
            # Nicotine <=1.4.1
            return downloads_file_1_4_1

        # Fall back to new file format
        return downloads_file_json

    def get_upload_list_file_name(self):

        data_dir = self.config.data_dir
        uploads_file_json = os.path.join(data_dir, 'uploads.json')

        return uploads_file_json

    @staticmethod
    def load_transfers_file(transfers_file):
        """ Loads a file of transfers in json format """

        if not os.path.isfile(transfers_file):
            return None

        with open(transfers_file, encoding="utf-8") as handle:
            return json.load(handle)

    @staticmethod
    def load_legacy_transfers_file(transfers_file):
        """ Loads a download queue file in pickle format (legacy) """

        if not os.path.isfile(transfers_file):
            return None

        with open(transfers_file, "rb") as handle:
            from pynicotine.utils import RestrictedUnpickler
            return RestrictedUnpickler(handle, encoding="utf-8").load()

    def load_transfers(self, transfer_type):

        load_func = self.load_transfers_file

        if transfer_type == "uploads":
            transfers_file = self.get_upload_list_file_name()
        else:
            transfers_file = self.get_download_queue_file_name()

        if transfer_type == "downloads" and not transfers_file.endswith("downloads.json"):
            load_func = self.load_legacy_transfers_file

        return load_file(transfers_file, load_func)

    def add_stored_transfers(self, transfer_type):

        transfers = self.load_transfers(transfer_type)

        if not transfers:
            return

        if transfer_type == "uploads":
            transfer_list = self.uploads
        else:
            transfer_list = self.downloads

        for i in transfers:
            loaded_status = current_byte_offset = bitrate = length = None
            size = 0

            try:
                loaded_status = str(i[3])
            except Exception:
                pass

            try:
                size = int(i[4])
            except Exception:
                pass

            try:
                current_byte_offset = int(i[5])
            except Exception:
                pass

            try:
                loaded_bitrate = i[6]

                if loaded_bitrate is not None:
                    bitrate = str(loaded_bitrate)
            except Exception:
                pass

            try:
                loaded_length = i[7]

                if loaded_length is not None:
                    length = str(loaded_length)
            except Exception:
                pass

            if loaded_status in ("Aborted", "Paused"):
                status = "Paused"

            elif loaded_status in ("Filtered", "Finished"):
                status = loaded_status

            elif current_byte_offset is not None and current_byte_offset >= size:
                status = "Finished"

            else:
                status = "User logged off"

            if transfer_type == "uploads" and status != "Finished":
                # Only finished uploads are supposed to be restored
                continue

            transfer_list.appendleft(
                Transfer(
                    user=i[0], filename=i[1], path=i[2], status=status,
                    size=size, current_byte_offset=current_byte_offset, bitrate=bitrate,
                    length=length
                )
            )

    def watch_stored_downloads(self):
        """ When logging in, we request to watch the status of our downloads """

        users = set()

        for i in self.downloads:
            users.add(i.user)

            if i.status in ("Aborted", "Paused"):
                i.status = "Paused"

            elif i.status in ("Filtered", "Finished"):
                continue

            else:
                i.status = "Getting status"

        for user in users:
            self.core.watch_user(user)

    """ Privileges """

    def set_privileged_users(self, user_list):
        for user in user_list:
            self.add_to_privileged(user)

    def add_to_privileged(self, user):
        self.privileged_users.add(user)

    def remove_from_privileged(self, user):
        if user in self.privileged_users:
            self.privileged_users.remove(user)

    def is_privileged(self, user):

        if not user:
            return False

        if user in self.privileged_users:
            return True

        return self.is_buddy_prioritized(user)

    def is_buddy_prioritized(self, user):

        if not user:
            return False

        for row in self.config.sections["server"]["userlist"]:
            if not row or not isinstance(row, list):
                continue

            if user == str(row[0]):
                # All users
                if self.config.sections["transfers"]["preferfriends"]:
                    return True

                # Only explicitly prioritized users
                try:
                    return bool(row[3])  # Prioritized column
                except IndexError:
                    return False

        return False

    """ File Actions """

    @staticmethod
    def get_file_size(filename):

        try:
            size = os.path.getsize(filename)
        except Exception:
            # file doesn't exist (remote files are always this)
            size = 0

        return size

    @staticmethod
    def close_file(file_handle, transfer):

        transfer.file = None

        if file_handle is None:
            return

        try:
            file_handle.close()

        except Exception as error:
            log.add_transfer("Failed to close file %(filename)s: %(error)s", {
                "filename": file_handle.name,
                "error": error
            })

    """ Limits """

    def _update_regular_limits(self):
        """ Sends the regular speed limits to the networking thread """

        uselimit = self.config.sections["transfers"]["uselimit"]
        uploadlimit = self.config.sections["transfers"]["uploadlimit"]
        limitby = self.config.sections["transfers"]["limitby"]

        self.queue.append(slskmessages.SetUploadLimit(uselimit, uploadlimit, limitby))
        self.queue.append(slskmessages.SetDownloadLimit(self.config.sections["transfers"]["downloadlimit"]))

    def _update_alt_limits(self):
        """ Sends the alternative speed limits to the networking thread """

        uselimit = True
        uploadlimit = self.config.sections["transfers"]["uploadlimitalt"]
        limitby = self.config.sections["transfers"]["limitby"]

        self.queue.append(slskmessages.SetUploadLimit(uselimit, uploadlimit, limitby))
        self.queue.append(slskmessages.SetDownloadLimit(self.config.sections["transfers"]["downloadlimitalt"]))

    def update_limits(self):

        if self.config.sections["transfers"]["usealtlimits"]:
            self._update_alt_limits()
            return

        self._update_regular_limits()

    def queue_limit_reached(self, user):

        queue_size_limit = self.config.sections["transfers"]["queuelimit"] * 1024 * 1024

        if not queue_size_limit:
            return False

        queue_size = 0

        for i in self.uploads:
            if i.user == user and i.status == "Queued":
                queue_size += i.size

                if queue_size >= queue_size_limit:
                    return True

        return False

    def file_limit_reached(self, user):

        file_limit = self.config.sections["transfers"]["filelimit"]

        if not file_limit:
            return False

        num_files = 0

        for i in self.uploads:
            if i.user == user and i.status == "Queued":
                num_files += 1

                if num_files >= file_limit:
                    return True

        return False

    def slot_limit_reached(self):

        upload_slot_limit = self.config.sections["transfers"]["uploadslots"]

        if upload_slot_limit <= 0:
            upload_slot_limit = 1

        num_in_progress = 0
        active_statuses = ("Getting status", "Establishing connection", "Transferring")

        for i in self.uploads:
            if i.status in active_statuses:
                num_in_progress += 1

                if num_in_progress >= upload_slot_limit:
                    return True

        return False

    def bandwidth_limit_reached(self):

        bandwidth_limit = self.config.sections["transfers"]["uploadbandwidth"] * 1024

        if not bandwidth_limit:
            return False

        bandwidth_sum = 0

        for i in self.uploads:
            if i.speed is not None and i.sock is not None:
                bandwidth_sum += i.speed

                if bandwidth_sum >= bandwidth_limit:
                    return True

        return False

    def allow_new_uploads(self):

        if self.config.sections["transfers"]["useupslots"]:
            # Limit by upload slots
            if self.slot_limit_reached():
                return False

        else:
            # Limit by maximum bandwidth
            if self.bandwidth_limit_reached():
                return False

        # No limits
        return True

    def file_is_upload_queued(self, user, filename):

        statuses = ("Queued", "Getting status", "Establishing connection", "Transferring")

        for i in self.uploads:
            if i.user == user and i.filename == filename and i.status in statuses:
                return True

        return False

    """ Network Events """

    def get_user_status(self, msg):
        """ We get a status of a user and if he's online, we request a file from him """

        download_statuses = ("Queued", "Getting status", "Establishing connection", "Too many files",
                             "Too many megabytes", "Pending shutdown.", "User logged off", "Connection closed by peer",
                             "Cannot connect", "Remote file error")

        upload_statuses = ("Getting status", "Establishing connection", "Disallowed extension",
                           "User logged off", "Cannot connect", "Cancelled")

        for i in reversed(self.downloads.copy()):
            if msg.user == i.user and (i.status in download_statuses or i.status.startswith("User limit of")):
                if msg.status <= 0:
                    i.status = "User logged off"
                    self.abort_transfer(i)
                    self.update_download(i)

                elif i.status in ("Getting status", "User logged off", "Connection closed by peer", "Cannot connect"):
                    self.get_file(i.user, i.filename, i.path, i)

        # We need a copy due to upload auto-clearing modifying the deque during iteration
        for i in reversed(self.uploads.copy()):
            if msg.user == i.user and i.status in upload_statuses:
                if msg.status <= 0:
                    i.status = "User logged off"
                    self.abort_transfer(i)

                    if not self.auto_clear_upload(i):
                        self.update_upload(i)

                elif i.status == "User logged off":
                    i.status = "Cancelled"

                    if not self.auto_clear_upload(i):
                        self.update_upload(i)

    def get_cant_connect_request(self, token):
        """ We can't connect to the user, either way (FileRequest, TransferRequest). """

        for i in self.downloads:
            if i.token == token:
                self._get_cant_connect_download(i)
                break

        for i in self.uploads:
            if i.token == token:
                self._get_cant_connect_upload(i)
                break

    def get_cant_connect_queue_file(self, username, filename):
        """ We can't connect to the user, either way (QueueUpload). """

        for i in self.downloads:
            if i.user == username and i.filename == filename:
                self._get_cant_connect_download(i)
                break

    def _get_cant_connect_download(self, i):

        i.status = "Cannot connect"
        i.token = None

        self.update_download(i)
        self.core.watch_user(i.user)

    def _get_cant_connect_upload(self, i):

        i.status = "Cannot connect"
        i.token = None
        i.queue_position = 0

        self.update_user_counter(i.user)
        self.update_upload(i)

        self.core.watch_user(i.user)
        self.check_upload_queue()

    def folder_contents_response(self, msg):
        """ When we got a contents of a folder, get all the files in it, but
        skip the files in subfolders """

        username = msg.init.target_user
        file_list = msg.list

        log.add_transfer("Received response for folder content request from user %s", username)

        for i in file_list:
            for directory in file_list[i]:
                if os.path.commonprefix([i, directory]) != directory:
                    continue

                files = file_list[i][directory][:]
                destination = self.get_folder_destination(username, directory)

                if self.config.sections["transfers"]["reverseorder"]:
                    files.sort(key=lambda x: x[1], reverse=True)

                log.add_transfer(
                    "Attempting to download files in folder %(folder)s for user %(user)s. "
                    + "Destination path: %(destination)s", {
                        "folder": directory,
                        "user": username,
                        "destination": destination
                    }
                )

                for file in files:
                    virtualpath = directory.rstrip('\\') + '\\' + file[1]
                    size = file[2]
                    h_bitrate, _bitrate, h_length, _length = get_result_bitrate_length(size, file[4])

                    self.get_file(
                        username, virtualpath, destination,
                        size=size, bitrate=h_bitrate, length=h_length)

                if directory in self.requested_folders.get(username, []):
                    del self.requested_folders[username][directory]

    def queue_upload(self, msg):
        """ Peer remotely queued a download (upload here). This is the modern replacement to
        a TransferRequest with direction 0 (download request). We will initiate the upload of
        the queued file later. """

        user = msg.init.target_user
        real_path = self.core.shares.virtual2real(msg.file)

        if not self.file_is_upload_queued(user, msg.file):

            limits = True

            if self.config.sections["transfers"]["friendsnolimits"]:
                friend = user in (i[0] for i in self.config.sections["server"]["userlist"])

                if friend:
                    limits = False

            ip_address, _port = msg.init.addr
            checkuser, reason = self.core.network_filter.check_user(user, ip_address)

            if not checkuser:
                self.queue.append(
                    slskmessages.UploadDenied(msg.init, msg.file, reason)
                )

            elif limits and self.queue_limit_reached(user):
                self.queue.append(
                    slskmessages.UploadDenied(msg.init, msg.file, "Too many megabytes")
                )

            elif limits and self.file_limit_reached(user):
                self.queue.append(
                    slskmessages.UploadDenied(msg.init, msg.file, "Too many files")
                )

            elif self.core.shares.file_is_shared(user, msg.file, real_path):
                newupload = Transfer(
                    user=user, filename=msg.file,
                    path=os.path.dirname(real_path), status="Queued",
                    size=self.get_file_size(real_path)
                )
                self.append_upload(user, msg.file, newupload)
                self.update_upload(newupload)

                self.core.pluginhandler.upload_queued_notification(user, msg.file, real_path)
                self.check_upload_queue()

            else:
                self.queue.append(
                    slskmessages.UploadDenied(msg.init, msg.file, "File not shared.")
                )

        log.add_transfer("QueueUpload request: User %(user)s, %(msg)s", {
            'user': user,
            'msg': str(vars(msg))
        })

    def transfer_request(self, msg):

        user = msg.init.target_user
        response = None

        if msg.direction == 1:
            log.add_transfer("Received upload request %(request)s for file %(filename)s from user %(user)s", {
                "request": msg.token,
                "filename": msg.file,
                "user": user
            })

            response = self.transfer_request_downloads(msg)

            log.add_transfer("Sending response to upload request %(request)s for file %(filename)s "
                             + "from user %(user)s: %(allowed)s", {
                                 "request": response.token,
                                 "filename": msg.file,
                                 "user": user,
                                 "allowed": response.allowed
                             })

        else:
            log.add_transfer("Received download request %(request)s for file %(filename)s from user %(user)s", {
                "request": msg.token,
                "filename": msg.file,
                "user": user
            })

            response = self.transfer_request_uploads(msg)

            log.add_transfer("Sending response to download request %(request)s for file %(filename)s "
                             + "from user %(user)s: %(allowed)s", {
                                 "request": response.token,
                                 "filename": msg.file,
                                 "user": user,
                                 "allowed": response.allowed
                             })

        self.core.send_message_to_peer(user, response)

    def transfer_request_downloads(self, msg):

        user = msg.init.target_user
        filename = msg.file
        size = msg.filesize

        cancel_reason = "Cancelled"
        accepted = True

        for i in self.downloads:
            if i.filename == filename and user == i.user:
                status = i.status

                if status == "Finished":
                    # SoulseekQt sends "Complete" as the reason for rejecting the download if it exists
                    cancel_reason = "Complete"
                    accepted = False
                    break

                if status in ("Paused", "Filtered"):
                    accepted = False
                    break

                # Remote peer is signaling a transfer is ready, attempting to download it

                # If the file is larger than 2GB, the SoulseekQt client seems to
                # send a malformed file size (0 bytes) in the TransferRequest response.
                # In that case, we rely on the cached, correct file size we received when
                # we initially added the download.

                if msg.filesize > 0:
                    i.size = size

                i.token = msg.token
                i.status = "Getting status"
                self.transfer_request_times[i] = time.time()

                self.update_download(i)
                response = slskmessages.TransferResponse(None, 1, token=i.token)
                return response

        # Check if download exists in our default download folder
        if self.get_existing_download_path(user, filename, "", size):
            cancel_reason = "Complete"
            accepted = False

        # If this file is not in your download queue, then it must be
        # a remotely initiated download and someone is manually uploading to you
        if accepted and self.can_upload(user):
            path = ""
            if self.config.sections["transfers"]["uploadsinsubdirs"]:
                parentdir = msg.file.replace('/', '\\').split('\\')[-2]
                path = os.path.join(self.config.sections["transfers"]["uploaddir"], user, parentdir)

            transfer = Transfer(
                user=user, filename=msg.file, path=path,
                status="Queued", size=msg.filesize, token=msg.token
            )
            self.downloads.appendleft(transfer)
            self.core.watch_user(user)

            response = slskmessages.TransferResponse(None, 0, reason="Queued", token=transfer.token)
            self.update_download(transfer)

        else:
            response = slskmessages.TransferResponse(None, 0, reason=cancel_reason, token=msg.token)
            log.add_transfer("Denied file request: User %(user)s, %(msg)s", {
                'user': user,
                'msg': str(vars(msg))
            })

        return response

    def transfer_request_uploads(self, msg):
        """ Remote peer is requesting to download a file through your upload queue.
        Note that the QueueUpload peer message has replaced this method of requesting
        a download in most clients. """

        response = self._transfer_request_uploads(msg)
        log.add_transfer("Legacy TransferRequest upload request: %(req)s Response: %(resp)s", {
            'req': str(vars(msg)),
            'resp': response
        })
        return response

    def _transfer_request_uploads(self, msg):

        # Is user allowed to download?
        user = msg.init.target_user
        ip_address, _port = msg.init.addr
        checkuser, reason = self.core.network_filter.check_user(user, ip_address)

        if not checkuser:
            return slskmessages.TransferResponse(None, 0, reason=reason, token=msg.token)

        # Do we actually share that file with the world?
        real_path = self.core.shares.virtual2real(msg.file)

        if not self.core.shares.file_is_shared(user, msg.file, real_path):
            return slskmessages.TransferResponse(None, 0, reason="File not shared.", token=msg.token)

        # Is that file already in the queue?
        if self.file_is_upload_queued(user, msg.file):
            return slskmessages.TransferResponse(None, 0, reason="Queued", token=msg.token)

        # Has user hit queue limit?
        limits = True

        if self.config.sections["transfers"]["friendsnolimits"]:
            friend = user in (i[0] for i in self.config.sections["server"]["userlist"])

            if friend:
                limits = False

        if limits and self.queue_limit_reached(user):
            return slskmessages.TransferResponse(None, 0, reason="Too many megabytes", token=msg.token)

        if limits and self.file_limit_reached(user):
            return slskmessages.TransferResponse(None, 0, reason="Too many files", token=msg.token)

        # All checks passed, user can queue file!
        self.core.pluginhandler.upload_queued_notification(user, msg.file, real_path)

        # Is user already downloading/negotiating a download?
        already_downloading = False
        active_statuses = ("Getting status", "Establishing connection", "Transferring")

        for i in self.uploads:
            if i.user != user:
                continue

            if i.status in active_statuses:
                already_downloading = True
                break

        if not self.allow_new_uploads() or already_downloading:

            response = slskmessages.TransferResponse(None, 0, reason="Queued", token=msg.token)
            newupload = Transfer(
                user=user, filename=msg.file,
                path=os.path.dirname(real_path), status="Queued",
                size=self.get_file_size(real_path)
            )
            self.append_upload(user, msg.file, newupload)
            self.update_upload(newupload)
            return response

        # All checks passed, starting a new upload.
        size = self.get_file_size(real_path)
        response = slskmessages.TransferResponse(None, 1, token=msg.token, filesize=size)

        transferobj = Transfer(
            user=user, filename=msg.file,
            path=os.path.dirname(real_path), status="Getting status",
            token=msg.token, size=size
        )

        self.transfer_request_times[transferobj] = time.time()
        self.append_upload(user, msg.file, transferobj)
        self.update_upload(transferobj)
        return response

    def transfer_response(self, msg):
        """ Received a response to the file request from the peer """

        log.add_transfer("Received response for transfer request %(request)s. Allowed: %(allowed)s. "
                         + "Reason: %(reason)s. Filesize: %(size)s", {
                             "request": msg.token,
                             "allowed": msg.allowed,
                             "reason": msg.reason,
                             "size": msg.filesize
                         })

        if msg.reason is not None:

            for i in self.uploads:
                if i.token != msg.token:
                    continue

                i.status = msg.reason
                i.token = None
                i.queue_position = 0

                self.update_upload(i)

                if i in self.transfer_request_times:
                    del self.transfer_request_times[i]

                self.update_user_counter(i.user)

                if msg.reason == "Complete":
                    # A complete download of this file already exists on the user's end
                    self.upload_finished(i)

                elif msg.reason in ("Cancelled", "Disallowed extension"):
                    self.auto_clear_upload(i)

                self.check_upload_queue()
                break

        else:
            for i in self.uploads:
                if i.token != msg.token:
                    continue

                i.status = "Establishing connection"
                self.core.send_message_to_peer(i.user, slskmessages.FileRequest(None, msg.token))
                self.update_upload(i)
                self.check_upload_queue()
                break
            else:
                log.add_transfer("Got unknown transfer response: %s", str(vars(msg)))

    def transfer_timeout(self, msg):

        transfer = msg.transfer

        if transfer.status == "Transferring":
            # Check if the transfer has started since the timeout callback was initiated
            return

        log.add_transfer("Transfer %(filename)s with request %(request)s for user %(user)s timed out", {
            "filename": transfer.filename,
            "request": transfer.token,
            "user": transfer.user
        })

        transfer.status = "Cannot connect"
        transfer.token = None

        self.core.watch_user(transfer.user)

        if transfer in self.downloads:
            self.update_download(transfer)

        elif transfer in self.uploads:
            transfer.queue_position = 0
            self.update_user_counter(transfer.user)
            self.update_upload(transfer)

        if transfer in self.transfer_request_times:
            del self.transfer_request_times[transfer]

        self.check_upload_queue()

    def file_error(self, msg):
        """ Networking thread encountered a local file error """

        for i in self.downloads + self.uploads:

            if i.sock != msg.sock:
                continue

            self.abort_transfer(i)
            i.status = "Local file error"

            log.add(_("I/O error: %s"), msg.strerror)

            if i in self.downloads:
                self.update_download(i)

            elif i in self.uploads:
                self.update_upload(i)
                self.check_upload_queue()

    def file_request(self, msg):
        """ Got an incoming file request. Could be an upload request or a
        request to get the file that was previously queued """

        token = msg.token

        log.add_transfer("Received file request %(request)s", {
            "request": token
        })

        for i in self.downloads:
            if token == i.token:
                self._file_request_download(msg, i)
                return

        for i in self.uploads:
            if token == i.token:
                self._file_request_upload(msg, i)
                return

        self.queue.append(slskmessages.FileConnClose(msg.init.sock))

    def _file_request_download(self, msg, i):

        log.add_transfer("Received file upload request %(request)s for file %(filename)s from user %(user)s", {
            "request": msg.token,
            "filename": i.filename,
            "user": i.user
        })

        incompletedir = self.config.sections["transfers"]["incompletedir"]
        needupdate = True

        if i.sock is None:
            i.sock = msg.init.sock
            i.token = None

            if i in self.transfer_request_times:
                del self.transfer_request_times[i]

            if not incompletedir:
                if i.path:
                    incompletedir = i.path
                else:
                    incompletedir = self.get_default_download_folder(i.user)

            try:
                if not os.access(incompletedir, os.F_OK):
                    os.makedirs(incompletedir)
                if not os.access(incompletedir, os.R_OK | os.W_OK | os.X_OK):
                    raise OSError("Download directory %s Permissions error.\nDir Permissions: %s" %
                                  (incompletedir, oct(os.stat(incompletedir)[stat.ST_MODE] & 0o777)))

            except OSError as error:
                log.add(_("OS error: %s"), error)
                self.download_folder_error(i, error)

            else:
                file_handle = None
                try:
                    from hashlib import md5
                    md5sum = md5()
                    md5sum.update((i.filename + i.user).encode('utf-8'))

                    base_name, extension = os.path.splitext(clean_file(i.filename.replace('/', '\\').split('\\')[-1]))
                    prefix = "INCOMPLETE" + md5sum.hexdigest()

                    # Ensure file name doesn't exceed 255 characters in length
                    incomplete_name = os.path.join(
                        incompletedir, prefix + base_name[:255 - len(prefix) - len(extension)] + extension)
                    file_handle = open(incomplete_name, 'ab+')

                    if self.config.sections["transfers"]["lock"]:
                        try:
                            import fcntl
                            try:
                                fcntl.lockf(file_handle, fcntl.LOCK_EX | fcntl.LOCK_NB)
                            except OSError as error:
                                log.add(_("Can't get an exclusive lock on file - I/O error: %s"), error)
                        except ImportError:
                            pass

                    file_handle.seek(0, 2)
                    offset = file_handle.tell()

                except OSError as error:
                    log.add(_("Download I/O error: %s"), error)

                    self.abort_transfer(i)
                    i.status = "Local file error"

                else:
                    i.file = file_handle
                    i.last_byte_offset = offset
                    i.queue_position = 0

                    self.core.statistics.append_stat_value("started_downloads", 1)
                    self.core.pluginhandler.download_started_notification(i.user, i.filename, incomplete_name)

                    if i.size > offset:
                        i.status = "Transferring"
                        i.legacy_attempt = False
                        self.queue.append(slskmessages.DownloadFile(i.sock, file_handle))
                        self.queue.append(slskmessages.FileOffset(msg.init, i.size, offset))

                        log.add_download(
                            _("Download started: user %(user)s, file %(file)s"), {
                                "user": i.user,
                                "file": "%s" % file_handle.name
                            }
                        )
                    else:
                        self.download_finished(file_handle, i)
                        needupdate = False

            if self.downloadsview:
                self.downloadsview.new_transfer_notification()

            if needupdate:
                self.update_download(i)

        else:
            log.add_transfer("Download error formally known as 'Unknown file request': %(req)s (%(user)s: %(file)s)", {
                'req': str(vars(msg)),
                'user': i.user,
                'file': i.filename
            })

            self.queue.append(slskmessages.FileConnClose(msg.init.sock))

    def _file_request_upload(self, msg, i):

        log.add_transfer("Received file download request %(request)s for file %(filename)s from user %(user)s", {
            "request": msg.token,
            "filename": i.filename,
            "user": i.user
        })
        needupdate = True

        if i.sock is None:
            i.sock = msg.init.sock
            i.token = None
            file_handle = None

            if i in self.transfer_request_times:
                del self.transfer_request_times[i]

            try:
                # Open File
                real_path = self.core.shares.virtual2real(i.filename)
                file_handle = open(real_path, "rb")
                offset = file_handle.tell()

            except OSError as error:
                log.add(_("Upload I/O error: %s"), error)

                self.abort_transfer(i)
                i.status = "Local file error"
                self.check_upload_queue()

            else:
                i.file = file_handle
                i.last_byte_offset = offset
                i.queue_position = 0

                self.core.statistics.append_stat_value("started_uploads", 1)
                self.core.pluginhandler.upload_started_notification(i.user, i.filename, real_path)

                if i.size > offset:
                    i.status = "Transferring"
                    self.queue.append(slskmessages.UploadFile(i.sock, file=file_handle, size=i.size))

                    log.add_upload(
                        _("Upload started: user %(user)s, IP address %(ip)s, file %(file)s"), {
                            "user": i.user,
                            "ip": self.core.protothread.user_addresses.get(i.user),
                            "file": i.filename
                        }
                    )
                else:
                    self.upload_finished(i, file_handle=file_handle)
                    needupdate = False

            if self.uploadsview:
                self.uploadsview.new_transfer_notification()

            if needupdate:
                self.update_upload(i)

        else:
            log.add_transfer("Upload error formally known as 'Unknown file request': %(req)s (%(user)s: %(file)s)", {
                'req': str(vars(msg)),
                'user': i.user,
                'file': i.filename
            })

            self.queue.append(slskmessages.FileConnClose(msg.init.sock))

    def upload_denied(self, msg):

        user = msg.init.target_user

        for i in self.downloads:
            if i.user != user or i.filename != msg.file:
                continue

            if i.status in ("Finished", "Paused"):
                # SoulseekQt also sends this message for finished downloads when unsharing files, ignore
                continue

            if msg.reason in ("File not shared.", "File not shared", "Remote file error") and not i.legacy_attempt:
                # The peer is possibly using an old client that doesn't support Unicode
                # (Soulseek NS). Attempt to request file name encoded as latin-1 once.

                log.add_transfer("User %(user)s responded with reason '%(reason)s' for download request %(filename)s. "
                                 "Attempting to request file as latin-1.", {
                                     "user": user,
                                     "reason": msg.reason,
                                     "filename": i.filename
                                 })

                self.abort_transfer(i)
                i.legacy_attempt = True
                self.get_file(i.user, i.filename, i.path, i)
                break

            if i.status == "Transferring":
                self.abort_transfer(i, reason=msg.reason)

            i.status = msg.reason
            self.update_download(i)

            log.add_transfer("Download request denied by user %(user)s for file %(filename)s. Reason: %(reason)s", {
                "user": user,
                "filename": i.filename,
                "reason": msg.reason
            })
            break

    def upload_failed(self, msg):

        user = msg.init.target_user

        for i in self.downloads:
            if i.user != user or i.filename != msg.file:
                continue

            if i.status in ("Finished", "Paused", "Download folder error", "Local file error", "User logged off"):
                # Check if there are more transfers with the same virtual path
                continue

            if not i.legacy_attempt:
                # Attempt to request file name encoded as latin-1 once

                self.abort_transfer(i)
                i.legacy_attempt = True
                self.get_file(i.user, i.filename, i.path, i)
                break

            # Already failed once previously, give up
            i.status = "Remote file error"
            self.update_download(i)

            log.add_transfer("Upload attempt by user %(user)s for file %(filename)s failed. Reason: %(reason)s", {
                "filename": i.filename,
                "user": user,
                "reason": "Remote file error"
            })
            break

    def file_download(self, msg):
        """ A file download is in progress """

        needupdate = True

        for i in self.downloads:

            if i.sock != msg.sock:
                continue

            try:
                if i in self.transfer_request_times:
                    del self.transfer_request_times[i]

                current_time = time.time()
                i.current_byte_offset = msg.file.tell()

                if i.start_time is None:
                    i.start_time = current_time

                i.status = "Transferring"
                oldelapsed = i.time_elapsed
                i.time_elapsed = current_time - i.start_time
                byte_difference = i.current_byte_offset - i.last_byte_offset

                if byte_difference:
                    self.core.statistics.append_stat_value("downloaded_size", byte_difference)

                if i.size > i.current_byte_offset:
                    if current_time > i.start_time and i.current_byte_offset > i.last_byte_offset:
                        i.speed = int(max(0, byte_difference // max(1, current_time - i.last_update)))

                        if i.speed <= 0:
                            i.time_left = "∞"
                        else:
                            i.time_left = human_length((i.size - i.current_byte_offset) / i.speed)

                    if oldelapsed == i.time_elapsed:
                        needupdate = False
                else:
                    self.download_finished(msg.file, i)
                    needupdate = False

                i.last_byte_offset = i.current_byte_offset
                i.last_update = current_time

            except OSError as error:
                log.add(_("Download I/O error: %s"), error)

                self.abort_transfer(i)
                i.status = "Local file error"

            if needupdate:
                self.update_download(i)

            break

    def file_upload(self, msg):
        """ A file upload is in progress """

        needupdate = True

        for i in self.uploads:

            if i.sock != msg.sock:
                continue

            if i in self.transfer_request_times:
                del self.transfer_request_times[i]

            current_time = time.time()
            i.current_byte_offset = msg.offset + msg.sentbytes

            if i.start_time is None:
                i.start_time = current_time

            i.status = "Transferring"
            oldelapsed = i.time_elapsed
            i.time_elapsed = current_time - i.start_time
            byte_difference = i.current_byte_offset - i.last_byte_offset

            if byte_difference:
                self.core.statistics.append_stat_value("uploaded_size", byte_difference)

            if i.size > i.current_byte_offset:
                if current_time > i.start_time and i.current_byte_offset > i.last_byte_offset:
                    i.speed = int(max(0, byte_difference // max(1, current_time - i.last_update)))

                    if i.speed <= 0:
                        i.time_left = "∞"
                    else:
                        i.time_left = human_length((i.size - i.current_byte_offset) / i.speed)

                if oldelapsed == i.time_elapsed:
                    needupdate = False

            elif i.speed is not None:
                # Inform the server about the last upload speed for this transfer
                self.upload_speed = i.speed
                self.queue.append(slskmessages.SendUploadSpeed(i.speed))

                i.speed = None
                i.time_left = ""

            i.last_byte_offset = i.current_byte_offset
            i.last_update = current_time

            if needupdate:
                self.update_upload(i)

            break

    def file_conn_close(self, sock):
        """ The remote peer has closed a file transfer connection """

        for i in self.downloads:
            if i.sock == sock:
                self._file_conn_close(i, "download")
                return

        # We need a copy due to upload auto-clearing modifying the deque during iteration
        for i in self.uploads.copy():
            if i.sock == sock:
                self._file_conn_close(i, "upload")
                return

    def _file_conn_close(self, i, transfer_type):

        if transfer_type == "download":
            self.abort_transfer(i)

            if i.status != "Finished":
                if self.user_logged_out(i.user):
                    i.status = "User logged off"
                else:
                    i.status = "Connection closed by peer"

            self.update_download(i)

        elif transfer_type == "upload":
            if i.current_byte_offset is not None and i.current_byte_offset >= i.size:
                # We finish the upload here in case the downloading peer has a slow/limited download
                # speed and finishes later than us
                self.upload_finished(i, file_handle=i.file)
                return

            if i.status == "Finished":
                return

            self.abort_transfer(i)

            if self.user_logged_out(i.user):
                i.status = "User logged off"
            else:
                i.status = "Cancelled"

                # Transfer ended abruptly. Tell the peer to re-queue the file. If the transfer was
                # intentionally cancelled, the peer should ignore this message.
                self.core.send_message_to_peer(i.user, slskmessages.UploadFailed(None, i.filename))

            if not self.auto_clear_upload(i):
                self.update_upload(i)

            self.check_upload_queue()

    def place_in_queue_request(self, msg):

        user = msg.init.target_user
        privileged_user = self.is_privileged(user)
        queue_position = 0
        transfer = None

        if self.config.sections["transfers"]["fifoqueue"]:
            for i in reversed(self.uploads):
                # Ignore non-queued files
                if i.status != "Queued":
                    continue

                if not privileged_user or self.is_privileged(i.user):
                    queue_position += 1

                # Stop counting on the matching file
                if i.user == user and i.filename == msg.file:
                    transfer = i
                    break

        else:
            num_queued_users = len(self.user_update_counters)

            for i in reversed(self.uploads):
                if i.user != user:
                    continue

                # Ignore non-queued files
                if i.status != "Queued":
                    continue

                queue_position += num_queued_users

                # Stop counting on the matching file
                if i.filename == msg.file:
                    transfer = i
                    break

        if queue_position > 0:
            self.queue.append(slskmessages.PlaceInQueue(msg.init, msg.file, queue_position))

        if transfer is None:
            return

        # Update queue position in our list of uploads
        transfer.queue_position = queue_position
        self.update_upload(transfer)

    def place_in_queue(self, msg):
        """ The peer tells us our place in queue for a particular transfer """

        username = msg.init.target_user
        filename = msg.filename

        for i in self.downloads:
            if i.user == username and i.filename == filename and i.status not in ("Finished", "Paused", "Filtered"):
                i.queue_position = msg.place
                self.update_download(i)
                return

    """ Transfer Actions """

    def get_folder(self, user, folder):
        self.core.send_message_to_peer(user, slskmessages.FolderContentsRequest(None, folder))

    def get_file(self, user, filename, path="", transfer=None, size=0, bitrate=None, length=None):

        path = clean_path(path, absolute=True)

        if transfer is None:
            for i in self.downloads:
                if i.user == user and i.filename == filename and i.path == path:
                    if i.status == "Finished":
                        # Duplicate finished download found, verify that it's still present on disk later
                        transfer = i
                        break

                    # Duplicate active/cancelled download found, stop here
                    return

            else:
                transfer = Transfer(
                    user=user, filename=filename, path=path,
                    status="Queued", size=size, bitrate=bitrate,
                    length=length
                )
                self.downloads.appendleft(transfer)
        else:
            transfer.filename = filename
            transfer.status = "Queued"

        self.core.watch_user(user)

        if self.config.sections["transfers"]["enablefilters"]:
            try:
                downloadregexp = re.compile(self.config.sections["transfers"]["downloadregexp"], re.I)

                if downloadregexp.search(filename) is not None:
                    log.add_transfer("Filtering: %s", filename)
                    self.abort_transfer(transfer)
                    # The string to be displayed on the GUI
                    transfer.status = "Filtered"

                    if self.auto_clear_download(transfer):
                        return

            except Exception:
                pass

        if self.user_logged_out(user):
            transfer.status = "User logged off"

        elif transfer.status != "Filtered":
            download_path = self.get_existing_download_path(user, filename, path, size)

            if download_path:
                transfer.status = "Finished"
                transfer.size = transfer.current_byte_offset = size

                log.add_transfer("File %s is already downloaded", download_path)

            else:
                log.add_transfer("Adding file %(filename)s from user %(user)s to download queue", {
                    "filename": filename,
                    "user": user
                })
                self.core.send_message_to_peer(
                    user, slskmessages.QueueUpload(None, filename, transfer.legacy_attempt))

        self.update_download(transfer)

    def push_file(self, user, filename, size, path="", transfer=None, bitrate=None, length=None, locally_queued=False):

        if not self.core.logged_in:
            return

        real_path = self.core.shares.virtual2real(filename)
        size_attempt = self.get_file_size(real_path)

        if size_attempt > 0:
            size = size_attempt

        if transfer is None:
            transfer = Transfer(
                user=user, filename=filename, path=path,
                status="Queued", size=size, bitrate=bitrate,
                length=length
            )
            self.append_upload(user, filename, transfer)
        else:
            transfer.filename = filename
            transfer.size = size
            transfer.status = "Queued"

        log.add_transfer(
            "Initializing upload request for file %(file)s to user %(user)s", {
                'file': filename,
                'user': user
            }
        )

        self.core.watch_user(user)

        if self.user_logged_out(user):
            transfer.status = "User logged off"

        elif not locally_queued:
            self.token = increment_token(self.token)
            transfer.token = self.token
            transfer.status = "Getting status"
            self.transfer_request_times[transfer] = time.time()

            log.add_transfer("Requesting to upload file %(filename)s with transfer "
                             + "request %(request)s to user %(user)s", {
                                 "filename": filename,
                                 "request": transfer.token,
                                 "user": user
                             })

            self.core.send_message_to_peer(
                user, slskmessages.TransferRequest(None, 1, transfer.token, filename, size, real_path))

        self.update_upload(transfer)

    def append_upload(self, user, filename, transferobj):

        previously_queued = False
        old_index = 0

        if self.is_privileged(user):
            transferobj.modifier = "privileged" if user in self.privileged_users else "prioritized"

        for i in self.uploads:
            if i.user == user and i.filename == filename:
                if i.status == "Queued":
                    # This upload was queued previously
                    # Use the previous queue position and time
                    transferobj.queue_position = i.queue_position
                    previously_queued = True

                if i in self.transfer_request_times:
                    del self.transfer_request_times[i]

                self.uploads.remove(i)

                if self.uploadsview:
                    self.uploadsview.remove_specific(i, True)

                break

            old_index += 1

        if previously_queued:
            self.uploads.insert(old_index, transferobj)
            return

        if user not in self.user_update_counters:
            self.update_user_counter(user)

        self.uploads.appendleft(transferobj)

    def can_upload(self, user):

        transfers = self.config.sections["transfers"]

        if transfers["remotedownloads"]:

            if transfers["uploadallowed"] == 0:
                # No One can sent files to you
                return False

            if transfers["uploadallowed"] == 1:
                # Everyone can sent files to you
                return True

            if (transfers["uploadallowed"] == 2
                    and user in (i[0] for i in self.config.sections["server"]["userlist"])):
                # Users in userlist
                return True

            if transfers["uploadallowed"] == 3:
                # Trusted buddies
                for row in self.config.sections["server"]["userlist"]:
                    if row[0] == user and row[4]:
                        return True

        return False

    def user_logged_out(self, user):
        """ Check if a user who previously queued a file has logged out since """

        if not self.core.logged_in:
            return True

        try:
            return self.core.user_statuses[user] <= 0

        except (KeyError, TypeError):
            return False

    def get_folder_destination(self, user, directory):

        # Check if a custom download location was specified
        if (user in self.requested_folders and directory in self.requested_folders[user]
                and self.requested_folders[user][directory]):
            download_location = self.requested_folders[user][directory]

        else:
            download_location = self.get_default_download_folder(user)

        # Get the last folder in directory path
        target_name = directory.rstrip('\\').split('\\')[-1]

        # Merge download path with target folder name
        destination = os.path.join(download_location, target_name)

        return destination

    def get_total_uploads_allowed(self):

        if self.config.sections["transfers"]["useupslots"]:
            maxupslots = self.config.sections["transfers"]["uploadslots"]

            if maxupslots <= 0:
                maxupslots = 1

            return maxupslots

        lstlen = sum(1 for i in self.uploads if i.sock is not None)

        if self.allow_new_uploads():
            return lstlen + 1

        return lstlen

    def get_upload_queue_size(self, username=None):

        if self.is_privileged(username):
            queue_size = 0

            for i in self.uploads:
                if i.status == "Queued" and self.is_privileged(i.user):
                    queue_size += 1

            return queue_size

        return sum(1 for i in self.uploads if i.status == "Queued")

    def get_default_download_folder(self, user):

        downloaddir = self.config.sections["transfers"]["downloaddir"]

        # Check if username subfolders should be created for downloads
        if self.config.sections["transfers"]["usernamesubfolders"]:
            try:
                downloaddir = os.path.join(downloaddir, clean_file(user))

                if not os.path.isdir(downloaddir):
                    os.makedirs(downloaddir)

            except Exception as error:
                log.add(_("Unable to save download to username subfolder, falling back "
                          "to default download folder. Error: %s") % error)

        return downloaddir

    def get_download_destination(self, user, virtual_path, target_path):
        """ Returns the download destination of a virtual file path """

        folder_path = target_path if target_path else self.get_default_download_folder(user)
        basename = clean_file(virtual_path.replace('/', '\\').split('\\')[-1])

        return folder_path, basename

    def get_existing_download_path(self, user, virtual_path, target_path, size):
        """ Returns the download path of a previous download, if available """

        folder, basename = self.get_download_destination(user, virtual_path, target_path)
        basename_root, extension = os.path.splitext(basename)
        download_path = os.path.join(folder, basename)
        counter = 1

        while os.path.isfile(download_path):
            if os.stat(download_path).st_size == size:
                # Found a previous download with a matching file size
                return download_path

            basename = basename_root + " (" + str(counter) + ")" + extension
            download_path = os.path.join(folder, basename)
            counter += 1

        return None

    @staticmethod
    def get_renamed(name):
        """ When a transfer is finished, we remove INCOMPLETE~ or INCOMPLETE
        prefix from the file's name.

        Checks if a file with the same name already exists, and adds a number
        to the file name if that's the case. """

        filename, extension = os.path.splitext(name)
        counter = 1

        while os.path.exists(name):
            name = filename + " (" + str(counter) + ")" + extension
            counter += 1

        return name

    def file_downloaded_actions(self, user, filepath):

        config = self.config.sections

        if config["notifications"]["notification_popup_file"]:
            self.core.notifications.new_text_notification(
                _("%(file)s downloaded from %(user)s") % {
                    'user': user,
                    'file': filepath.rsplit(os.sep, 1)[1]
                },
                title=_("File downloaded")
            )

        if config["transfers"]["afterfinish"]:
            if not execute_command(config["transfers"]["afterfinish"], filepath):
                log.add(_("Trouble executing '%s'"), config["transfers"]["afterfinish"])
            else:
                log.add(_("Executed: %s"), config["transfers"]["afterfinish"])

    def folder_downloaded_actions(self, user, folderpath):

        # walk through downloads and break if any file in the same folder exists, else execute
        for i in self.downloads:
            if i.status not in ("Finished", "Paused", "Filtered") and i.path == folderpath:
                return

        config = self.config.sections

        if not folderpath:
            return

        if config["notifications"]["notification_popup_folder"]:
            self.core.notifications.new_text_notification(
                _("%(folder)s downloaded from %(user)s") % {
                    'user': user,
                    'folder': folderpath
                },
                title=_("Folder downloaded")
            )

        if config["transfers"]["afterfolder"]:
            if not execute_command(config["transfers"]["afterfolder"], folderpath):
                log.add(_("Trouble executing on folder: %s"), config["transfers"]["afterfolder"])
            else:
                log.add(_("Executed on folder: %s"), config["transfers"]["afterfolder"])

    def download_folder_error(self, transfer, error):

        self.abort_transfer(transfer)
        transfer.status = "Download folder error"

        self.core.notifications.new_text_notification(
            _("OS error: %s") % error, title=_("Download folder error"))

    def download_finished(self, file, i):

        self.close_file(file, i)

        folder, basename = self.get_download_destination(i.user, i.filename, i.path)
        newname = self.get_renamed(os.path.join(folder, basename))

        try:
            if not os.access(folder, os.F_OK):
                os.makedirs(folder)

            import shutil
            shutil.move(file.name, newname)

        except OSError as inst:
            log.add(
                _("Couldn't move '%(tempfile)s' to '%(file)s': %(error)s"), {
                    'tempfile': "%s" % file.name,
                    'file': newname,
                    'error': inst
                }
            )
            self.download_folder_error(i, inst)
            self.update_download(i)
            return

        i.status = "Finished"
        i.current_byte_offset = i.size
        i.speed = None
        i.time_left = ""
        i.sock = None

        self.core.statistics.append_stat_value("completed_downloads", 1)

        # Attempt to show notification and execute commands
        self.file_downloaded_actions(i.user, newname)
        self.folder_downloaded_actions(i.user, i.path)

        if self.downloadsview:
            # Main tab highlight (bright)
            self.downloadsview.new_transfer_notification(finished=True)

        # Attempt to autoclear this download, if configured
        if not self.auto_clear_download(i, force_save=True):
            self.update_download(i, force_save=True)

        self.core.pluginhandler.download_finished_notification(i.user, i.filename, newname)

        log.add_download(
            _("Download finished: user %(user)s, file %(file)s"), {
                'user': i.user,
                'file': i.filename
            }
        )

    def upload_finished(self, i, file_handle=None):

        self.close_file(file_handle, i)

        i.status = "Finished"
        i.current_byte_offset = i.size
        i.speed = None
        i.time_left = ""
        i.sock = None

        self.update_user_counter(i.user)

        log.add_upload(
            _("Upload finished: user %(user)s, IP address %(ip)s, file %(file)s"), {
                'user': i.user,
                'ip': self.core.protothread.user_addresses.get(i.user),
                'file': i.filename
            }
        )

        self.core.statistics.append_stat_value("completed_uploads", 1)

        # Autoclear this upload
        if not self.auto_clear_upload(i, force_save=True):
            self.update_upload(i, force_save=True)

        real_path = self.core.shares.virtual2real(i.filename)
        self.core.pluginhandler.upload_finished_notification(i.user, i.filename, real_path)

        self.check_upload_queue()

    def auto_clear_download(self, transfer, force_save=False):

        if self.config.sections["transfers"]["autoclear_downloads"]:
            self.downloads.remove(transfer)
            self.save_transfers("downloads", force_save)

            if self.downloadsview:
                self.downloadsview.remove_specific(transfer, True)

            return True

        return False

    def auto_clear_upload(self, transfer, force_save=False):

        if self.config.sections["transfers"]["autoclear_uploads"]:
            self.uploads.remove(transfer)
            self.save_transfers("uploads", force_save)

            if self.uploadsview:
                self.uploadsview.remove_specific(transfer, True)

            return True

        return False

    def update_download(self, transfer, force_save=False):

        self.save_transfers("downloads", force_save)

        if self.downloadsview:
            self.downloadsview.update(transfer)

    def update_upload(self, transfer, force_save=False):

        self.save_transfers("uploads", force_save)

        if self.uploadsview:
            self.uploadsview.update(transfer)

    def _check_transfer_timeouts(self):

        while True:
            current_time = time.time()

            if self.transfer_request_times:
                for transfer, start_time in self.transfer_request_times.copy().items():
                    # When our port is closed, certain clients can take up to ~30 seconds before they
                    # initiate a 'F' connection, since they only send an indirect connection request after
                    # attempting to connect to our port for a certain time period.
                    # Known clients: Nicotine+ 2.2.0 - 3.2.0, 2 s; Soulseek NS, ~20 s; soulseeX, ~30 s.
                    # To account for potential delays while initializing the connection, add 15 seconds
                    # to the timeout value.

                    if (current_time - start_time) >= 45:
                        self.network_callback([slskmessages.TransferTimeout(transfer)])

            if self.core.protothread.exit.wait(1):
                # Event set, we're exiting
                return

    def _check_upload_queue_timer(self):

        while True:
            self.network_callback([slskmessages.CheckUploadQueue()])

            if self.core.protothread.exit.wait(10):
                # Event set, we're exiting
                return

    def _check_download_queue_timer(self):

        self.download_queue_timer_count = -1

        while True:
            self.download_queue_timer_count += 1
            self.network_callback([slskmessages.CheckDownloadQueue()])

            if self.core.protothread.exit.wait(180):
                # Event set, we're exiting
                return

    # Find failed or stuck downloads and attempt to queue them.
    # Also ask for the queue position of downloads.
    def check_download_queue(self):

        statuslist_failed = ("Cannot connect", "Connection closed by peer", "Local file error", "Remote file error")
        statuslist_limited = ("Too many files", "Too many megabytes")
        reset_count = False

        for transfer in reversed(self.downloads):
            status = transfer.status

            if (self.download_queue_timer_count >= 4
                    and (status in statuslist_limited or status.startswith("User limit of"))):
                # Re-queue limited downloads every 12 minutes

                log.add_transfer("Re-queuing file %(filename)s from user %(user)s in download queue", {
                    "filename": transfer.filename,
                    "user": transfer.user
                })
                reset_count = True

                self.abort_transfer(transfer)
                self.get_file(transfer.user, transfer.filename, transfer.path, transfer)

            if status in statuslist_failed:
                # Retry failed downloads every 3 minutes

                self.abort_transfer(transfer)
                self.get_file(transfer.user, transfer.filename, transfer.path, transfer)

            if status == "Queued":
                # Request queue position every 3 minutes

                self.core.send_message_to_peer(
                    transfer.user,
                    slskmessages.PlaceInQueueRequest(None, transfer.filename, transfer.legacy_attempt)
                )

        if reset_count:
            self.download_queue_timer_count = 0

    def get_upload_candidate(self):
        """ Retrieve a suitable queued transfer for uploading.
        Round Robin: Get the first queued item from the oldest user
        FIFO: Get the first queued item in the list """

        round_robin_queue = not self.config.sections["transfers"]["fifoqueue"]
        active_statuses = ("Getting status", "Establishing connection", "Transferring")
        privileged_queue = False

        first_queued_transfers = OrderedDict()
        queued_users = {}
        uploading_users = set()

        for i in reversed(self.uploads):
            if i.status == "Queued":
                user = i.user

                if user not in first_queued_transfers and user not in uploading_users:
                    first_queued_transfers[user] = i

                if user in queued_users:
                    continue

                privileged = self.is_privileged(user)
                queued_users[user] = privileged

                if privileged:
                    privileged_queue = True

            elif i.status in active_statuses:
                # We're currently uploading a file to the user
                user = i.user

                uploading_users.add(user)

                if user in first_queued_transfers:
                    del first_queued_transfers[user]

        oldest_time = None
        target_user = None

        if not round_robin_queue:
            # skip the looping below (except the cleanup) and get the first
            # user of the highest priority we saw above
            for user in first_queued_transfers:
                if privileged_queue and not queued_users[user]:
                    continue

                target_user = user
                break

        for user, update_time in list(self.user_update_counters.items()):
            if user not in queued_users:
                del self.user_update_counters[user]
                continue

            if not round_robin_queue or user in uploading_users:
                continue

            if privileged_queue and not queued_users[user]:
                continue

            if not oldest_time:
                oldest_time = update_time + 1

            if update_time < oldest_time:
                target_user = user
                oldest_time = update_time

        if not target_user:
            return None

        return first_queued_transfers[target_user]

    def check_upload_queue(self):
        """ Find next file to upload """

        while True:
            if not self.uploads:
                # No uploads exist
                return

            if not self.allow_new_uploads():
                return

            upload_candidate = self.get_upload_candidate()

            if upload_candidate is None:
                return

            user = upload_candidate.user

            log.add_transfer(
                "Checked upload queue, attempting to upload file %(file)s to user %(user)s", {
                    'file': upload_candidate.filename,
                    'user': user
                }
            )

            if self.user_logged_out(user):
                upload_candidate.status = "User logged off"
                self.abort_transfer(upload_candidate)

                if not self.auto_clear_upload(upload_candidate):
                    self.update_upload(upload_candidate)

                # Check queue again
                continue

            self.push_file(
                user=user, filename=upload_candidate.filename, size=upload_candidate.size, transfer=upload_candidate
            )
            return

    def update_user_counter(self, user):
        """ Called when an upload associated with a user has changed. The user update counter
        is used by the Round Robin queue system to determine which user has waited the longest
        since their last download. """

        self.user_update_counter += 1
        self.user_update_counters[user] = self.user_update_counter

    def ban_user(self, user, ban_message=None):
        """ Ban a user, cancel all the user's uploads, send a 'Banned'
        message via the transfers, and clear the transfers from the
        uploads list. """

        if ban_message:
            banmsg = "Banned (%s)" % ban_message
        elif self.config.sections["transfers"]["usecustomban"]:
            banmsg = "Banned (%s)" % self.config.sections["transfers"]["customban"]
        else:
            banmsg = "Banned"

        for upload in self.uploads.copy():
            if upload.user != user:
                continue

            self.abort_transfer(upload, reason=banmsg, send_fail_message=True)

            if self.uploadsview:
                self.uploadsview.remove_specific(upload)

        self.core.network_filter.ban_user(user)

    def retry_download(self, transfer):

        if transfer.status in ("Transferring", "Finished"):
            return

        user = transfer.user

        if self.user_logged_out(user):
            transfer.status = "User logged off"
            self.abort_transfer(transfer)
            self.update_download(transfer)
            return

        self.abort_transfer(transfer)
        self.get_file(user, transfer.filename, transfer.path, transfer)

    def retry_upload(self, transfer):

        active_statuses = ["Getting status", "Establishing connection", "Transferring"]

        if transfer.status in active_statuses + ["Finished"]:
            # Don't retry active or finished uploads
            return

        user = transfer.user

        for i in self.uploads:
            if i.user != user:
                continue

            if i.status in active_statuses:
                # User already has an active upload, queue the retry attempt
                if transfer.status != "Queued":
                    transfer.status = "Queued"
                    self.update_upload(transfer)
                return

        if self.user_logged_out(user):
            transfer.status = "User logged off"
            self.abort_transfer(transfer)

            if not self.auto_clear_upload(transfer):
                self.update_upload(transfer)
            return

        self.push_file(user, transfer.filename, transfer.size, transfer.path, transfer=transfer)

    def abort_transfer(self, transfer, reason="Cancelled", send_fail_message=False):

        transfer.legacy_attempt = False
        transfer.token = None
        transfer.speed = None
        transfer.queue_position = 0
        transfer.time_left = ""

        if transfer.sock is not None:
            self.queue.append(slskmessages.FileConnClose(transfer.sock))
            transfer.sock = None

        if transfer in self.transfer_request_times:
            del self.transfer_request_times[transfer]

        if transfer.file is not None:
            self.close_file(transfer.file, transfer)

            if transfer in self.uploads:
                self.update_user_counter(transfer.user)
                self.check_upload_queue()
                log.add_upload(
                    _("Upload aborted, user %(user)s file %(file)s"), {
                        "user": transfer.user,
                        "file": transfer.filename
                    }
                )
            else:
                log.add_download(
                    _("Download aborted, user %(user)s file %(file)s"), {
                        "user": transfer.user,
                        "file": transfer.filename
                    }
                )

        elif send_fail_message and transfer in self.uploads and transfer.status == "Queued":
            self.core.send_message_to_peer(
                transfer.user, slskmessages.UploadDenied(None, file=transfer.filename, reason=reason))

    """ Filters """

    def update_download_filters(self):

        proccessedfilters = []
        outfilter = "(\\\\("
        failed = {}
        download_filters = sorted(self.config.sections["transfers"]["downloadfilters"])
        # Get Filters from config file and check their escaped status
        # Test if they are valid regular expressions and save error messages

        for item in download_filters:
            dfilter, escaped = item
            if escaped:
                dfilter = re.escape(dfilter)
                dfilter = dfilter.replace("\\*", ".*")
            else:
                # Avoid "Nothing to repeat" error
                dfilter = dfilter.replace("*", "\\*").replace("+", "\\+")

            try:
                re.compile("(" + dfilter + ")")
                outfilter += dfilter
                proccessedfilters.append(dfilter)

            except Exception as error:
                failed[dfilter] = error

            proccessedfilters.append(dfilter)

            if item is not download_filters[-1]:
                outfilter += "|"

        # Crop trailing pipes
        while outfilter[-1] == "|":
            outfilter = outfilter[:-1]

        outfilter += ")$)"
        try:
            re.compile(outfilter)
            self.config.sections["transfers"]["downloadregexp"] = outfilter
            # Send error messages for each failed filter to log window
            if failed:
                errors = ""

                for dfilter, error in failed.items():
                    errors += "Filter: %s Error: %s " % (dfilter, error)

                log.add(_("Error: %(num)d Download filters failed! %(error)s "), {'num': len(failed), 'error': errors})

        except Exception as error:
            # Strange that individual filters _and_ the composite filter both fail
            log.add(_("Error: Download Filter failed! Verify your filters. Reason: %s"), error)
            self.config.sections["transfers"]["downloadregexp"] = ""

    """ Exit """

    def abort_transfers(self):
        """ Stop all transfers on disconnect/shutdown """

        for i in self.downloads:
            if i.status not in ("Finished", "Filtered", "Paused"):
                self.abort_transfer(i)
                i.status = "User logged off"
                self.update_download(i)

        for i in self.uploads.copy():
            if i.status != "Finished":
                self.uploads.remove(i)

                if self.uploadsview:
                    self.uploadsview.remove_specific(i, True)

        self.privileged_users.clear()
        self.requested_folders.clear()
        self.transfer_request_times.clear()
        self.user_update_counters.clear()

    def get_downloads(self):
        """ Get a list of downloads """
        return [[i.user, i.filename, i.path, i.status, i.size, i.current_byte_offset, i.bitrate, i.length]
                for i in reversed(self.downloads)]

    def get_uploads(self):
        """ Get a list of finished uploads """
        return [[i.user, i.filename, i.path, i.status, i.size, i.current_byte_offset, i.bitrate, i.length]
                for i in reversed(self.uploads) if i.status == "Finished"]

    def save_downloads_callback(self, filename):
        json.dump(self.get_downloads(), filename, ensure_ascii=False)

    def save_uploads_callback(self, filename):
        json.dump(self.get_uploads(), filename, ensure_ascii=False)

    def save_transfers(self, transfer_type, force_save=False):
        """ Save list of transfers """

        if not self.allow_saving_transfers:
            # Don't save if transfers didn't load properly!
            return

        self.config.create_data_folder()
        current_time = time.time()

        if not force_save and (current_time - self.last_save_times[transfer_type]) < 15:
            # Save list of transfers to file every 15 seconds
            return

        if transfer_type == "uploads":
            transfers_file = self.uploads_file_name
            callback = self.save_uploads_callback
        else:
            transfers_file = self.downloads_file_name
            callback = self.save_downloads_callback

        write_file_and_backup(transfers_file, callback)
        self.last_save_times[transfer_type] = current_time

    def server_disconnect(self):

        self.abort_transfers()

        if self.downloadsview:
            self.downloadsview.server_disconnect()

        if self.uploadsview:
            self.uploadsview.server_disconnect()

    def quit(self):

        self.save_transfers("downloads", force_save=True)
        self.save_transfers("uploads", force_save=True)


class Statistics:

    def __init__(self, config, ui_callback=None):

        self.config = config
        self.ui_callback = ui_callback
        self.session_stats = {}

        if hasattr(ui_callback, "statistics"):
            self.ui_callback = ui_callback.statistics

        for stat_id in self.config.defaults["statistics"]:
            self.session_stats[stat_id] = 0

    def append_stat_value(self, stat_id, stat_value):

        self.session_stats[stat_id] += stat_value
        self.config.sections["statistics"][stat_id] += stat_value
        self.update_ui(stat_id)

    def update_ui(self, stat_id):

        if self.ui_callback:
            stat_value = self.session_stats[stat_id]
            self.ui_callback.update_stat_value(stat_id, stat_value)

    def reset_stats(self):

        for stat_id in self.config.defaults["statistics"]:
            self.session_stats[stat_id] = 0
            self.config.sections["statistics"][stat_id] = 0

            self.update_ui(stat_id)



# COPYRIGHT (C) 2020-2022 Nicotine+ Team
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from os.path import commonprefix

from gi.repository import Gtk

from pynicotine import slskmessages
from pynicotine.config import config
from pynicotine.gtkgui.widgets.accelerator import Accelerator
from pynicotine.logfacility import log
from pynicotine.utils import add_alias
from pynicotine.utils import get_alias
from pynicotine.utils import is_alias
from pynicotine.utils import unalias


""" Text Entry-related """


class ChatEntry:
    """ Custom text entry with support for chat commands and completions """

    def __init__(self, frame, entry, completion, entity, message_class, send_message, command_list, is_chatroom=False):

        self.frame = frame
        self.entry = entry
        self.completion = completion
        self.entity = entity
        self.message_class = message_class
        self.send_message = send_message
        self.command_list = command_list
        self.is_chatroom = is_chatroom

        entry.connect("activate", self.on_enter)
        Accelerator("<Shift>Tab", entry, self.on_tab_complete_accelerator, True)
        Accelerator("Tab", entry, self.on_tab_complete_accelerator)

        # Emoji Picker
        try:
            self.entry.set_property("show-emoji-icon", True)

        except TypeError:
            # GTK version not supported
            pass

        # Spell Check
        if config.sections["ui"]["spellcheck"]:
            if not self.frame.spell_checker:
                self.frame.init_spell_checker()

            if self.frame.spell_checker:
                from gi.repository import Gspell  # pylint:disable=no-name-in-module
                spell_buffer = Gspell.EntryBuffer.get_from_gtk_entry_buffer(entry.get_buffer())
                spell_buffer.set_spell_checker(self.frame.spell_checker)
                spell_view = Gspell.Entry.get_from_gtk_entry(entry)
                spell_view.set_inline_spell_checking(True)

    def on_enter(self, *_args):

        if not self.frame.np.logged_in:
            return

        text = self.entry.get_text()

        if not text:
            return

        if is_alias(text):
            alias_text = get_alias(text)

            if not alias_text:
                log.add(_('Alias "%s" returned nothing'), text)
                return

            text = alias_text

        is_double_slash_cmd = text.startswith("//")
        is_single_slash_cmd = (text.startswith("/") and not is_double_slash_cmd)

        if not is_single_slash_cmd or text.startswith("/me"):
            # Regular chat message (/me is sent as plain text)

            self.entry.set_text("")

            if is_double_slash_cmd:
                # Remove first slash and send the rest of the command as plain text
                text = text[1:]

            self.send_message(self.entity, text)
            return

        cmd_split = text.split(maxsplit=1)
        cmd = cmd_split[0]

        if cmd + " " not in self.command_list:
            log.add(_("Command %s is not recognized"), cmd)
            return

        # Clear chat entry
        self.entry.set_text("")

        if len(cmd_split) == 2:
            args = arg_self = cmd_split[1]
        else:
            args = ""
            arg_self = "" if self.is_chatroom else self.entity

        if cmd in ("/alias", "/al"):
            parent = self.frame.np.chatrooms if self.is_chatroom else self.frame.np.privatechats
            parent.echo_message(self.entity, add_alias(args))

            if config.sections["words"]["aliases"]:
                self.frame.update_completions()

        elif cmd in ("/unalias", "/un"):
            parent = self.frame.np.chatrooms if self.is_chatroom else self.frame.np.privatechats
            parent.echo_message(self.entity, unalias(args))

            if config.sections["words"]["aliases"]:
                self.frame.update_completions()

        elif cmd in ("/w", "/whois", "/info"):
            if arg_self:
                self.frame.np.userinfo.request_user_info(arg_self)
                self.frame.change_main_page("userinfo")

        elif cmd in ("/b", "/browse"):
            if arg_self:
                self.frame.np.userbrowse.browse_user(arg_self)
                self.frame.change_main_page("userbrowse")

        elif cmd == "/ip":
            if arg_self:
                self.frame.np.request_ip_address(arg_self)

        elif cmd == "/pm":
            if args:
                self.frame.np.privatechats.show_user(args)
                self.frame.change_main_page("private")

        elif cmd in ("/m", "/msg"):
            if args:
                args_split = args.split(" ", maxsplit=1)
                user = args_split[0]
                msg = None

                if len(args_split) == 2:
                    msg = args_split[1]

                if msg:
                    self.frame.np.privatechats.show_user(user)
                    self.frame.np.privatechats.send_message(user, msg)
                    self.frame.change_main_page("private")

        elif cmd in ("/s", "/search"):
            if args:
                self.frame.np.search.do_search(args, "global")
                self.frame.change_main_page("search")

        elif cmd in ("/us", "/usearch"):
            args_split = args.split(" ", maxsplit=1)

            if len(args_split) == 2:
                self.frame.np.search.do_search(args_split[1], "user", user=args_split[0])
                self.frame.change_main_page("search")

        elif cmd in ("/rs", "/rsearch"):
            if args:
                self.frame.np.search.do_search(args, "rooms")
                self.frame.change_main_page("search")

        elif cmd in ("/bs", "/bsearch"):
            if args:
                self.frame.np.search.do_search(args, "buddies")
                self.frame.change_main_page("search")

        elif cmd in ("/j", "/join"):
            if args:
                self.frame.np.queue.append(slskmessages.JoinRoom(args))

        elif cmd in ("/l", "/leave", "/p", "/part"):
            if args:
                self.frame.np.queue.append(slskmessages.LeaveRoom(args))
            else:
                self.frame.np.queue.append(slskmessages.LeaveRoom(self.entity))

        elif cmd in ("/ad", "/add", "/buddy"):
            if args:
                self.frame.np.userlist.add_user(args)

        elif cmd in ("/rem", "/unbuddy"):
            if args:
                self.frame.np.userlist.remove_user(args)

        elif cmd == "/ban":
            if args:
                self.frame.np.network_filter.ban_user(args)

        elif cmd == "/ignore":
            if args:
                self.frame.np.network_filter.ignore_user(args)

        elif cmd == "/ignoreip":
            if args:
                self.frame.np.network_filter.ignore_ip(args)

        elif cmd == "/unban":
            if args:
                self.frame.np.network_filter.unban_user(args)

        elif cmd == "/unignore":
            if args:
                self.frame.np.network_filter.unignore_user(args)

        elif cmd == "/ctcpversion":
            if arg_self:
                self.frame.np.privatechats.show_user(arg_self)
                self.frame.np.privatechats.send_message(arg_self, self.frame.np.privatechats.CTCP_VERSION)

        elif cmd in ("/clear", "/cl"):
            if self.is_chatroom:
                parent = self.frame.chatrooms.pages[self.entity]
            else:
                parent = self.frame.privatechat.pages[self.entity]

            parent.chat_textview.clear()

        elif cmd in ("/a", "/away"):
            self.frame.on_away()

        elif cmd in ("/q", "/quit", "/exit"):
            self.frame.np.quit()

        elif cmd in ("/c", "/close"):
            self.frame.privatechat.pages[self.entity].on_close()

        elif cmd == "/now":
            self.frame.np.now_playing.display_now_playing(
                callback=lambda np_message: self.send_message(self.entity, np_message))

        elif cmd == "/rescan":
            self.frame.np.shares.rescan_shares()

        elif cmd == "/toggle":
            if args:
                self.frame.np.pluginhandler.toggle_plugin(args)

        elif self.is_chatroom:
            self.frame.np.pluginhandler.trigger_public_command_event(self.entity, cmd[1:], args)

        elif not self.is_chatroom:
            self.frame.np.pluginhandler.trigger_private_command_event(self.entity, cmd[1:], args)

    def on_tab_complete_accelerator(self, widget, state, backwards=False):
        """ Tab and Shift+Tab: tab complete chat """
        return self.completion.on_tab_complete_accelerator(widget, state, backwards)


class ChatCompletion:

    def __init__(self):

        self.completion_list = []
        self.completion_iters = {}

        self.midwaycompletion = False  # True if the user just used tab completion
        self.completions = {}  # Holds temp. information about tab completoin

        self.entry = None
        self.entry_changed_handler = None
        self.entry_completion = None
        self.model = Gtk.ListStore(str)

        self.column_numbers = list(range(self.model.get_n_columns()))

    def create_entry_completion(self):

        self.entry_completion = Gtk.EntryCompletion(model=self.model)
        self.entry_completion.set_text_column(0)
        self.entry_completion.set_match_func(self.entry_completion_find_match)
        self.entry_completion.connect("match-selected", self.entry_completion_found_match)

    def set_entry(self, entry):

        if self.entry is not None:
            self.entry.set_completion(None)
            self.entry.disconnect(self.entry_changed_handler)

        # Reusing an existing GtkEntryCompletion object after unsetting it doesn't work well
        self.create_entry_completion()
        entry.set_completion(self.entry_completion)

        self.entry = entry
        self.entry_changed_handler = entry.connect("changed", self.on_entry_changed)

    def add_completion(self, item):

        if not config.sections["words"]["tab"]:
            return

        if item in self.completion_list:
            return

        self.completion_list.append(item)

        if config.sections["words"]["dropdown"]:
            self.completion_iters[item] = self.model.insert_with_valuesv(-1, self.column_numbers, [item])

    def get_completion(self, part, completion_list):

        matches = self.get_completions(part, completion_list)

        if not matches:
            return None, 0

        if len(matches) == 1:
            return matches[0], 1

        return commonprefix([x.lower() for x in matches]), 0

    @staticmethod
    def get_completions(part, completion_list):

        lowerpart = part.lower()
        matches = [x for x in completion_list if x.lower().startswith(lowerpart) and len(x) >= len(part)]
        return matches

    def remove_completion(self, item):

        if not config.sections["words"]["tab"]:
            return

        if item not in self.completion_list:
            return

        self.completion_list.remove(item)

        if not config.sections["words"]["dropdown"]:
            return

        iterator = self.completion_iters[item]
        self.model.remove(iterator)
        del self.completion_iters[item]

    def set_completion_list(self, completion_list):

        if self.entry_completion is None:
            return

        config_words = config.sections["words"]

        self.entry_completion.set_popup_single_match(not config_words["onematch"])
        self.entry_completion.set_minimum_key_length(config_words["characters"])
        self.entry_completion.set_inline_completion(False)

        self.model.clear()
        self.completion_iters.clear()

        if not config_words["tab"]:
            return

        if completion_list is None:
            completion_list = []

        self.completion_list = completion_list

        if not config_words["dropdown"]:
            self.entry_completion.set_popup_completion(False)
            return

        for word in completion_list:
            word = str(word)
            self.completion_iters[word] = self.model.insert_with_valuesv(-1, self.column_numbers, [word])

        self.entry_completion.set_popup_completion(True)

    def entry_completion_find_match(self, _completion, entry_text, iterator):

        if not entry_text:
            return False

        # Get word to the left of current position
        if " " in entry_text:
            i = self.entry.get_position()
            split_key = entry_text[:i].split(" ")[-1]
        else:
            split_key = entry_text

        if not split_key or len(split_key) < config.sections["words"]["characters"]:
            return False

        # Case-insensitive matching
        item_text = self.model.get_value(iterator, 0).lower()

        if item_text.startswith(split_key) and item_text != split_key:
            return True

        return False

    def entry_completion_found_match(self, _completion, model, iterator):

        current_text = self.entry.get_text()
        completion_value = model.get_value(iterator, 0)

        # if more than a word has been typed, we throw away the
        # one to the left of our current position because we want
        # to replace it with the matching word

        if " " in current_text:
            i = self.entry.get_position()
            prefix = " ".join(current_text[:i].split(" ")[:-1])
            suffix = " ".join(current_text[i:].split(" "))

            # add the matching word
            new_text = "%s %s%s" % (prefix, completion_value, suffix)
            # set back the whole text
            self.entry.set_text(new_text)
            # move the cursor at the end
            self.entry.set_position(len(prefix) + len(completion_value) + 1)
        else:
            self.entry.set_text(completion_value)
            self.entry.set_position(-1)

        # stop the event propagation
        return True

    def on_entry_changed(self, *_args):
        # If the entry was modified, and we don't block the handler, we're no longer completing
        self.midwaycompletion = False

    def on_tab_complete_accelerator(self, _widget, _state, backwards=False):
        """ Tab and Shift+Tab: tab complete chat """

        if not config.sections["words"]["tab"]:
            return False

        text = self.entry.get_text()

        if not text:
            return False

        # "Hello there Miss<tab> how are you doing"
        # "0  3  6  9  12 15      18 21 24 27 30 33
        #   1  4  7  10 13      16 19 22 25 28 31
        #    2  5  8  11 14      17 20 23 26 29 32
        #
        # i = 16
        # text = Miss
        # preix = 12
        i = self.entry.get_position()
        text = text[:i].split(" ")[-1]
        preix = i - len(text)

        if not config.sections["words"]["cycle"]:
            completion, single = self.get_completion(text, self.completion_list)
            if completion:
                if single and i == len(text) and not text.startswith("/"):
                    completion += ": "
                self.entry.delete_text(preix, i)
                self.entry.insert_text(completion, preix)
                self.entry.set_position(preix + len(completion))

            return True

        if not self.midwaycompletion:
            self.completions['completions'] = self.get_completions(text, self.completion_list)

            if self.completions['completions']:
                self.midwaycompletion = True
                self.completions['currentindex'] = -1
                currentnick = text
        else:
            currentnick = self.completions['completions'][self.completions['currentindex']]

        if self.midwaycompletion:
            # We're still completing, block handler to avoid modifying midwaycompletion value
            with self.entry.handler_block(self.entry_changed_handler):
                self.entry.delete_text(i - len(currentnick), i)
                direction = 1  # Forward cycle

                if backwards:
                    direction = -1  # Backward cycle

                self.completions['currentindex'] = ((self.completions['currentindex'] + direction) %
                                                    len(self.completions['completions']))

                newnick = self.completions['completions'][self.completions['currentindex']]
                self.entry.insert_text(newnick, preix)
                self.entry.set_position(preix + len(newnick))

        return True


class CompletionEntry:

    def __init__(self, entry, model, column=0):

        self.entry = entry
        self.model = model
        self.column = column

        completion = Gtk.EntryCompletion(inline_completion=True, inline_selection=True,
                                         popup_single_match=False, model=model)
        completion.set_text_column(column)
        completion.set_match_func(self.entry_completion_find_match)
        entry.set_completion(completion)

    def entry_completion_find_match(self, _completion, entry_text, iterator):

        if not entry_text:
            return False

        item_text = self.model.get_value(iterator, self.column)

        if not item_text:
            return False

        if item_text.lower().startswith(entry_text.lower()):
            return True

        return False


class TextSearchBar:

    def __init__(self, textview, search_bar, entry, controller_widget=None, focus_widget=None):

        self.textview = textview
        self.search_bar = search_bar
        self.entry = entry
        self.focus_widget = focus_widget or textview

        self.search_bar.connect_entry(self.entry)

        self.entry.connect("activate", self.on_search_next_match)
        self.entry.connect("search-changed", self.on_search_changed)

        self.entry.connect("previous-match", self.on_search_previous_match)
        self.entry.connect("next-match", self.on_search_next_match)

        if not controller_widget:
            controller_widget = textview

        Accelerator("<Primary>f", controller_widget, self.on_show_search_accelerator)

        for widget in (controller_widget, entry):
            Accelerator("Escape", widget, self.on_hide_search_accelerator)

    def on_search_match(self, search_type, restarted=False):

        buffer = self.textview.get_buffer()
        query = self.entry.get_text()

        self.textview.emit("select-all", False)

        if search_type == "typing":
            start, end = buffer.get_bounds()
            iterator = start
        else:
            current = buffer.get_mark("insert")
            iterator = buffer.get_iter_at_mark(current)

        if search_type == "previous":
            match = iterator.backward_search(
                query, Gtk.TextSearchFlags.TEXT_ONLY | Gtk.TextSearchFlags.CASE_INSENSITIVE, limit=None)
        else:
            match = iterator.forward_search(
                query, Gtk.TextSearchFlags.TEXT_ONLY | Gtk.TextSearchFlags.CASE_INSENSITIVE, limit=None)

        if match is not None and len(match) == 2:
            match_start, match_end = match

            if search_type == "previous":
                buffer.place_cursor(match_start)
                buffer.select_range(match_start, match_end)
            else:
                buffer.place_cursor(match_end)
                buffer.select_range(match_end, match_start)

            self.textview.scroll_to_iter(match_start, 0, False, 0.5, 0.5)

        elif not restarted and search_type != "typing":
            start, end = buffer.get_bounds()

            if search_type == "previous":
                buffer.place_cursor(end)
            elif search_type == "next":
                buffer.place_cursor(start)

            self.on_search_match(search_type, restarted=True)

    def on_search_changed(self, *_args):
        self.on_search_match(search_type="typing")

    def on_search_previous_match(self, *_args):
        self.on_search_match(search_type="previous")

    def on_search_next_match(self, *_args):
        self.on_search_match(search_type="next")

    def on_hide_search_accelerator(self, *_args):
        """ Escape: hide search bar """

        self.hide_search_bar()
        return True

    def on_show_search_accelerator(self, *_args):
        """ Ctrl+F: show search bar """

        self.show_search_bar()
        return True

    def show_search_bar(self):
        self.search_bar.set_search_mode(True)
        self.entry.grab_focus()

    def hide_search_bar(self):
        self.search_bar.set_search_mode(False)
        self.focus_widget.grab_focus()



# COPYRIGHT (C) 2020-2021 Nicotine+ Team
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import time

from pynicotine.config import config


class Logger:

    def __init__(self):

        self.listeners = set()
        self.log_levels = None
        self.file_name = "debug_" + str(int(time.time()))

    @staticmethod
    def set_msg_prefix(level, msg):

        if level == "download":
            prefix = "Download"
        elif level == "upload":
            prefix = "Upload"
        elif level == "search":
            prefix = "Search"
        elif level == "chat":
            prefix = "Chat"
        elif level == "connection":
            prefix = "Conn"
        elif level == "message":
            prefix = "Msg"
        elif level == "transfer":
            prefix = "Transfer"
        elif level == "miscellaneous":
            prefix = "Misc"
        else:
            prefix = ""

        if prefix:
            msg = "[%s] %s" % (prefix, msg)

        return str(msg)

    def add(self, msg, msg_args=None, level=None):

        if self.log_levels:
            levels = self.log_levels
        else:
            levels = config.sections["logging"].get("debugmodes", "")

        # Important messages are always visible
        if level and not level.startswith("important") and level not in levels:
            return

        if not msg_args and level == "message":
            # Compile message contents
            msg = "%s %s" % (msg.__class__, self.contents(msg))

        msg = self.set_msg_prefix(level, msg)

        if msg_args:
            msg = msg % msg_args

        timestamp_format = config.sections["logging"].get("log_timestamp", "%Y-%m-%d %H:%M:%S")

        if config.sections["logging"].get("debug_file_output", False):
            folder = config.sections["logging"]["debuglogsdir"]

            self.write_log(folder, self.file_name, msg, timestamp_format=timestamp_format)

        for callback in self.listeners:
            try:
                callback(timestamp_format, msg, level)
            except Exception as error:
                try:
                    print("Callback on %s failed: %s %s\n%s" % (callback, level, msg, error))
                except OSError:
                    # stdout is gone
                    pass

    def add_download(self, msg, msg_args=None):
        self.log_transfer(msg, msg_args)
        self.add(msg, msg_args=msg_args, level="download")

    def add_upload(self, msg, msg_args=None):
        self.log_transfer(msg, msg_args)
        self.add(msg, msg_args=msg_args, level="upload")

    def add_search(self, msg, msg_args=None):
        self.add(msg, msg_args=msg_args, level="search")

    def add_chat(self, msg, msg_args=None):
        self.add(msg, msg_args=msg_args, level="chat")

    def add_conn(self, msg, msg_args=None):
        self.add(msg, msg_args=msg_args, level="connection")

    def add_msg_contents(self, msg, msg_args=None):
        self.add(msg, msg_args=msg_args, level="message")

    def add_transfer(self, msg, msg_args=None):
        self.add(msg, msg_args=msg_args, level="transfer")

    def add_debug(self, msg, msg_args=None):
        self.add(msg, msg_args=msg_args, level="miscellaneous")

    def add_important_error(self, msg, msg_args=None):
        self.add(msg, msg_args=msg_args, level="important_error")

    def add_important_info(self, msg, msg_args=None):
        self.add(msg, msg_args=msg_args, level="important_info")

    @staticmethod
    def contents(obj):
        """ Returns variables for object, for debug output """
        try:
            return {s: getattr(obj, s) for s in obj.__slots__ if hasattr(obj, s)}
        except AttributeError:
            return vars(obj)

    def add_listener(self, callback):
        self.listeners.add(callback)

    def remove_listener(self, callback):
        try:
            self.listeners.remove(callback)
        except KeyError:
            self.add("Failed to remove listener %s, does not exist.", callback)

    def log_transfer(self, msg, msg_args=None):

        if not config.sections["logging"]["transfers"]:
            return

        folder = config.sections["logging"]["transferslogsdir"]
        timestamp_format = config.sections["logging"]["log_timestamp"]

        if msg_args:
            msg = msg % msg_args

        self.write_log(folder, "transfers", msg, timestamp_format=timestamp_format)

    def write_log(self, logsdir, filename, msg, timestamp=None, timestamp_format="%Y-%m-%d %H:%M:%S"):

        try:
            from pynicotine.utils import clean_file
            filename = clean_file(filename) + ".log"
            oldumask = os.umask(0o077)

            if not os.path.exists(logsdir):
                os.makedirs(logsdir)

            from pynicotine.utils import get_path
            get_path(logsdir, filename, self.write_log_callback, (oldumask, timestamp, timestamp_format, msg))

        except Exception as error:
            self.add(_("Couldn't write to log file \"%(filename)s\": %(error)s") %
                     {"filename": filename, "error": error})

    @staticmethod
    def write_log_callback(path, data):

        oldumask, timestamp, timestamp_format, msg = data

        with open(path, 'ab', 0) as logfile:
            os.umask(oldumask)

            text = "%s %s\n" % (time.strftime(timestamp_format, time.localtime(timestamp)), msg)
            logfile.write(text.encode('utf-8', 'replace'))


class Console:

    def __init__(self, logger):
        logger.add_listener(self.console_logger)

    @staticmethod
    def console_logger(timestamp_format, msg, _level):
        try:
            print("[" + time.strftime(timestamp_format) + "] " + msg)
        except OSError:
            # stdout is gone
            pass


log = Logger()
Console(log)



# COPYRIGHT (C) 2020-2021 Nicotine+ Team
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

from collections import deque

from pynicotine.config import config
from pynicotine.shares import Shares

SHARES_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), ".sharedfiles")


class SharesTest(unittest.TestCase):

    def setUp(self):

        config.data_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "dbs")
        config.filename = os.path.join(config.data_dir, "temp_config")

        config.load_config()

    def test_shares_scan(self):
        """ Test a full shares scan """

        config.sections["transfers"]["shared"] = [("Shares", SHARES_DIR)]
        config.sections["transfers"]["rescanonstartup"] = False

        shares = Shares(None, config, deque(), init_shares=False)
        shares.rescan_shares(use_thread=False)

        # Verify that modification time was saved for shares folder
        self.assertIn(SHARES_DIR, list(shares.share_dbs["mtimes"]))

        # Verify that shared files were added
        self.assertIn(['dummy_file', 0, None, None], shares.share_dbs["files"]["Shares"])
        self.assertIn(['nicotinetestdata.mp3', 80919, (128, 0), 5], shares.share_dbs["files"]["Shares"])
        self.assertIn(['vbr44.mp3', 36609, (32, 1), 9], shares.share_dbs["files"]["Shares"])  # 1 = VBR

        # Verify that expected folder is empty
        self.assertEqual(len(shares.share_dbs["files"]["Shares\\folder2"]), 0)

        # Verify that search index was updated
        word_index = shares.share_dbs["wordindex"]
        nicotinetestdata_indexes = list(word_index["nicotinetestdata"])
        ogg_indexes = list(word_index["ogg"])

        self.assertEqual(set(word_index), set(
            ['nicotinetestdata', 'ogg', 'mp3', 'shares', 'file', 'dummy', 'folder1',
             'folder2', 'nothing', 'something', 'test', 'vbr44']
        ))
        self.assertEqual(len(nicotinetestdata_indexes), 2)
        self.assertEqual(len(ogg_indexes), 1)

        # File ID associated with word "ogg" should return our nicotinetestdata.ogg file
        self.assertIn(ogg_indexes[0], nicotinetestdata_indexes)
        self.assertEqual(shares.share_dbs["fileindex"][str(ogg_indexes[0])][0], 'Shares\\nicotinetestdata.ogg')

        shares.close_shares(shares.share_dbs)

    def test_hidden_file_folder_scan(self):
        """ Test that hidden files and folders are excluded """

        config.sections["transfers"]["shared"] = [("Shares", SHARES_DIR)]
        config.sections["transfers"]["rescanonstartup"] = False

        shares = Shares(None, config, deque(), init_shares=False)
        shares.rescan_shares(use_thread=False)

        # Check folders
        mtimes = list(shares.share_dbs["mtimes"])

        self.assertNotIn(os.path.join(SHARES_DIR, ".abc"), mtimes)
        self.assertNotIn(os.path.join(SHARES_DIR, ".xyz"), mtimes)
        self.assertIn(os.path.join(SHARES_DIR, "folder1"), mtimes)
        self.assertIn(os.path.join(SHARES_DIR, "folder2"), mtimes)
        self.assertNotIn(os.path.join(SHARES_DIR, "folder2", ".poof"), mtimes)
        self.assertIn(os.path.join(SHARES_DIR, "folder2", "test"), mtimes)
        self.assertIn(os.path.join(SHARES_DIR, "something"), mtimes)

        # Check files
        files = shares.share_dbs["files"]["Shares"]

        self.assertNotIn([".abc_file", 0, None, None], files)
        self.assertNotIn([".hidden_file", 0, None, None], files)
        self.assertNotIn([".xyz_file", 0, None, None], files)
        self.assertIn(["dummy_file", 0, None, None], files)
        self.assertEqual(len(files), 4)

        shares.close_shares(shares.share_dbs)



# COPYRIGHT (C) 2020-2022 Nicotine+ Team
# COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
# COPYRIGHT (C) 2008-2009 Quinox <quinox@users.sf.net>
# COPYRIGHT (C) 2006-2009 Daelstorm <daelstorm@gmail.com>
# COPYRIGHT (C) 2003-2004 Hyriand <hyriand@thegraveyard.org>
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from gi.repository import Gdk
from gi.repository import Gtk

from pynicotine.gtkgui.widgets.dialogs import option_dialog
from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
from pynicotine.gtkgui.widgets.theme import get_icon
from pynicotine.gtkgui.widgets.theme import get_status_icon
from pynicotine.config import config


""" Icon Notebook """


class TabLabel(Gtk.Box):

    def __init__(self, label="", full_text="", close_button_visible=False, close_callback=None):

        Gtk.Box.__init__(self, hexpand=False, visible=True)

        self.highlighted = False
        self.mentioned = False
        self.centered = False
        self.gesture_click = None

        if Gtk.get_major_version() == 4:
            self.eventbox = Gtk.Box()
        else:
            self.eventbox = Gtk.EventBox(visible=True)

        self.box = Gtk.Box(spacing=6, visible=True)

        self.label = Gtk.Label(halign=Gtk.Align.START, hexpand=True, visible=True)
        self.full_text = full_text
        self.set_text(label)

        self.close_button = None
        self.close_button_visible = close_button_visible
        self.close_callback = close_callback

        self.start_icon = Gtk.Image()
        self.start_icon_data = None
        self.start_icon.hide()

        self.end_icon = Gtk.Image()
        self.end_icon_data = None
        self.end_icon.hide()

        self._pack_children()

    def _remove_tab_label(self):

        if self.eventbox.get_parent() is None:
            return

        for widget in (self.start_icon, self.label, self.end_icon):
            self.box.remove(widget)

        self.eventbox.remove(self.box)
        self.remove(self.eventbox)

    def _add_close_button(self):

        if self.close_button is not None:
            return

        if not self.close_button_visible:
            return

        if Gtk.get_major_version() == 4:
            self.close_button = Gtk.Button.new_from_icon_name("window-close-symbolic")

            # GTK 4 workaround to prevent notebook tabs from being activated when pressing close button
            gesture_click = Gtk.GestureClick()
            gesture_click.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
            gesture_click.connect(
                "pressed", lambda controller, *args: controller.set_state(Gtk.EventSequenceState.CLAIMED))
            self.close_button.add_controller(gesture_click)

        else:
            self.close_button = Gtk.Button.new_from_icon_name("window-close-symbolic", Gtk.IconSize.BUTTON)

        self.close_button.get_style_context().add_class("flat")
        self.close_button.set_tooltip_text(_("Close tab"))
        self.close_button.show()
        self.add(self.close_button)

        if self.close_callback is not None:
            self.close_button.connect("clicked", self.close_callback)

    def _remove_close_button(self):

        if self.close_button is not None:
            self.remove(self.close_button)
            self.close_button = None

    def _pack_children(self):

        self._remove_tab_label()
        self._remove_close_button()

        if sys.platform == "darwin":
            # Left align close button on macOS
            self._add_close_button()

        self.add(self.eventbox)
        self.eventbox.add(self.box)

        if self.centered:
            self.set_halign(Gtk.Align.CENTER)
        else:
            self.set_halign(Gtk.Align.FILL)

        self.box.add(self.start_icon)
        self.box.add(self.label)
        self.box.add(self.end_icon)

        if sys.platform != "darwin":
            self._add_close_button()

    def _set_text_color(self, color):

        color_rgba = Gdk.RGBA()

        if color_rgba.parse(color):
            from html import escape
            self.label.set_markup("<span foreground=\"%s\">%s</span>" % (color, escape(self.text)))
            return

        self.label.set_text("%s" % self.text)

    def set_centered(self, centered):
        self.centered = centered
        self._pack_children()

    def set_close_button_visibility(self, visible):

        self.close_button_visible = visible

        if visible:
            self._add_close_button()
            return

        self._remove_close_button()

    def request_hilite(self, mentioned=False):

        self.highlighted = True

        # Chat mentions have priority over normal notifications
        if not self.mentioned:
            self.mentioned = mentioned

        color = config.sections["ui"]["tab_default"]

        if config.sections["notifications"]["notification_tab_colors"]:
            if self.mentioned:
                color = config.sections["ui"]["tab_hilite"]
            else:
                color = config.sections["ui"]["tab_changed"]

        self._set_text_color(color)

        if self.mentioned:
            icon_data = get_icon("hilite")
        else:
            icon_data = get_icon("hilite3")

        if icon_data is self.end_icon_data:
            return

        self.end_icon_data = icon_data
        self.end_icon.set_property("gicon", icon_data)
        self.end_icon.show()

    def remove_hilite(self):

        self.highlighted = False
        self.mentioned = False

        self._set_text_color(config.sections["ui"]["tab_default"])

        self.end_icon_data = None
        self.end_icon.set_property("gicon", None)
        self.end_icon.hide()

    def set_status_icon(self, status):

        icon_data = get_status_icon(status)

        if icon_data is self.start_icon_data:
            return

        self.start_icon_data = icon_data
        self.start_icon.set_property("gicon", icon_data)
        self.start_icon.set_visible(config.sections["ui"]["tab_status_icons"])

    def set_start_icon_name(self, icon_name):
        self.start_icon.set_property("icon-name", icon_name)
        self.start_icon.show()

    def set_text(self, text):

        self.text = text

        if self.highlighted:
            self.request_hilite()
            return

        self._set_text_color(config.sections["ui"]["tab_default"])

    def get_text(self):
        return self.label.get_text()


class IconNotebook:
    """ This class extends the functionality of a Gtk.Notebook widget. On top of what a Gtk.Notebook provides:
    - Icons on tabs
    - Context (right-click) menus for tabs
    - Dropdown menu for unread tabs
    """

    def __init__(self, frame, notebook, page_id):

        self.notebook = notebook
        self.notebook.set_show_tabs(False)
        self.notebook.connect("page-removed", self.on_remove_page)
        self.notebook.connect("switch-page", self.on_switch_page)

        self.frame = frame
        self.page_id = page_id
        self.unread_button = Gtk.MenuButton(
            tooltip_text=_("Unread Tabs"),
            halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER, visible=False
        )
        self.pages = {}

        if Gtk.get_major_version() == 4:
            self.window = self.notebook.get_root()
            self.unread_button.set_has_frame(False)
            self.unread_button.set_icon_name("emblem-important-symbolic")

            # GTK 4 workaround to prevent notebook tabs from being activated when pressing close button
            controllers = self.notebook.observe_controllers()

            for num in range(controllers.get_n_items()):
                item = controllers.get_item(num)

                if isinstance(item, Gtk.GestureClick):
                    item.set_propagation_phase(Gtk.PropagationPhase.BUBBLE)
                    break

        else:
            self.window = self.notebook.get_toplevel()
            self.unread_button.set_image(Gtk.Image(icon_name="emblem-important-symbolic"))

        style_context = self.unread_button.get_style_context()
        for style_class in ("circular", "flat"):
            style_context.add_class(style_class)

        self.notebook.set_action_widget(self.unread_button, Gtk.PackType.END)

        self.popup_menu_unread = PopupMenu(self.frame, connect_events=False)
        self.unread_button.set_menu_model(self.popup_menu_unread.model)
        self.unread_pages = []

        self.notebook.popup_enable()

    """ Tabs """

    def get_labels(self, page):
        tab_label = self.notebook.get_tab_label(page)
        menu_label = self.notebook.get_menu_label(page)

        return tab_label, menu_label

    def get_tab_label_inner(self, page):
        return self.notebook.get_tab_label(page).eventbox

    def set_tab_closers(self):

        for i in range(self.notebook.get_n_pages()):
            page = self.notebook.get_nth_page(i)
            tab_label, _menu_label = self.get_labels(page)
            tab_label.set_close_button_visibility(config.sections["ui"]["tabclosers"])

    def set_tab_text_colors(self):

        for i in range(self.notebook.get_n_pages()):
            page = self.notebook.get_nth_page(i)
            tab_label, _menu_label = self.get_labels(page)
            tab_label.set_text(tab_label.get_text())

    def append_page(self, page, text, close_callback=None, full_text=None, user=None):

        if full_text is None:
            full_text = text

        label_tab = TabLabel(text, full_text, config.sections["ui"]["tabclosers"], close_callback)
        label_tab.set_tooltip_text(full_text)

        if Gtk.get_major_version() == 4:
            label_tab.gesture_click = Gtk.GestureClick()
            label_tab.add_controller(label_tab.gesture_click)
        else:
            label_tab.gesture_click = Gtk.GestureMultiPress(widget=label_tab)

        label_tab.gesture_click.set_button(Gdk.BUTTON_MIDDLE)
        label_tab.gesture_click.connect("pressed", label_tab.close_callback, page)

        # menu for all tabs
        label_tab_menu = TabLabel(text)

        self.notebook.append_page_menu(page, label_tab, label_tab_menu)

        if user is not None:
            status = 0

            if user in self.frame.np.user_statuses:
                status = self.frame.np.user_statuses[user] or 0

            self.set_user_status(page, text, status)

        self.notebook.set_tab_reorderable(page, True)
        self.notebook.set_show_tabs(True)

    def remove_page(self, page):

        self.notebook.remove_page(self.page_num(page))

        self.remove_unread_page(page)

        if self.notebook.get_n_pages() == 0:
            self.notebook.set_show_tabs(False)

    def remove_all_pages_response(self, dialog, response_id, _data):

        dialog.destroy()

        if response_id == 2:
            for i in reversed(range(self.notebook.get_n_pages())):
                page = self.notebook.get_nth_page(i)
                tab_label, _menu_label = self.get_labels(page)
                tab_label.close_callback(dialog)

    def remove_all_pages(self):

        option_dialog(
            parent=self.window,
            title=_('Close All Tabs?'),
            message=_('Do you really want to close all tabs?'),
            callback=self.remove_all_pages_response
        )

    def get_current_page(self):
        return self.notebook.get_current_page()

    def set_current_page(self, page_num):
        return self.notebook.set_current_page(page_num)

    def set_tab_pos(self, pos):
        self.notebook.set_tab_pos(pos)

    def get_n_pages(self):
        return self.notebook.get_n_pages()

    def get_nth_page(self, page_num):
        return self.notebook.get_nth_page(page_num)

    def page_num(self, page):
        return self.notebook.page_num(page)

    def next_page(self):
        return self.notebook.next_page()

    def prev_page(self):
        return self.notebook.prev_page()

    """ Tab Highlights """

    def request_tab_hilite(self, page, mentioned=False):

        page_active = (self.get_nth_page(self.get_current_page()) == page)

        if self.frame.current_page_id != self.page_id or not page_active:
            # Highlight top-level tab
            self.frame.request_tab_hilite(self.page_id, mentioned)

        if page_active:
            return

        tab_label, menu_label = self.get_labels(page)
        tab_label.request_hilite(mentioned)
        menu_label.request_hilite(mentioned)

        self.append_unread_page(page)

    def remove_tab_hilite(self, page):

        tab_label, menu_label = self.get_labels(page)
        tab_label.remove_hilite()
        menu_label.remove_hilite()

        self.remove_unread_page(page)

    def append_unread_page(self, page):

        if page in self.unread_pages:
            return

        self.unread_pages.append(page)
        self.update_unread_pages_menu()
        self.unread_button.show()

    def remove_unread_page(self, page):

        if page in self.unread_pages:
            self.unread_pages.remove(page)
            self.update_unread_pages_menu()

        if not self.unread_pages:
            self.unread_button.hide()
            self.frame.remove_tab_hilite(self.page_id)

    def set_unread_page(self, _action, _state, page):
        self.notebook.set_current_page(self.page_num(page))

    def update_unread_pages_menu(self):

        self.popup_menu_unread.clear()

        for page in self.unread_pages:
            tab_label, _menu_label = self.get_labels(page)
            self.popup_menu_unread.add_items(
                ("#" + tab_label.get_text(), self.set_unread_page, page)
            )

        self.popup_menu_unread.update_model()

    """ Tab User Status """

    def set_user_status(self, page, user, status):

        if status is None:
            return

        if status == 1:
            status_text = _("Away")
        elif status == 2:
            status_text = _("Online")
        else:
            status_text = _("Offline")

        if not config.sections["ui"]["tab_status_icons"]:
            tab_text = "%s (%s)" % (user[:15], status_text)
        else:
            tab_text = user

        tab_label, menu_label = self.get_labels(page)

        tab_label.set_status_icon(status)
        menu_label.set_status_icon(status)

        tab_label.set_text(tab_text)
        menu_label.set_text(tab_text)

        tab_label.set_tooltip_text("%s (%s)" % (user, status_text))

    """ Signals """

    def on_remove_page(self, _notebook, new_page, _page_num):
        self.remove_unread_page(new_page)

    def on_switch_page(self, _notebook, new_page, _page_num):

        # Hide container widget on previous page for a performance boost
        current_page = self.get_nth_page(self.get_current_page())

        if Gtk.get_major_version() == 4:
            current_page.get_first_child().hide()
            new_page.get_first_child().show()
        else:
            current_page.get_children()[0].hide()
            new_page.get_children()[0].show()

        # Dismiss tab highlight
        self.remove_tab_hilite(new_page)

    def on_tab_popup(self, widget, page):
        # Dummy implementation
        pass
